from .conditional import Conditional
from .enumeration import Enum, EnumItem
from .factor import Factor, FactorEncryptMethod, FactorIndexGroup, FactorType
from .pipeline import Pipeline, PipelineStage, PipelineTriggerType, PipelineUnit
from .pipeline_action import AggregateArithmetic, AggregateArithmeticHolder, DeleteTopicActionType, FindBy, \
	FromFactor, FromTopic, MemoryWriter, PipelineAction, PipelineActionType, ReadTopicActionType, SystemActionType, \
	ToFactor, ToTopic, WriteTopicActionType
from .pipeline_action_delete import DeleteRowAction, DeleteRowsAction, DeleteTopicAction
from .pipeline_action_read import ExistsAction, ReadFactorAction, ReadFactorsAction, ReadRowAction, ReadRowsAction, \
	ReadTopicAction
from .pipeline_action_system import AlarmAction, AlarmActionSeverity, CopyToMemoryAction, WriteToExternalAction
from .pipeline_action_write import InsertRowAction, MappingFactor, MappingRow, MergeRowAction, WriteFactorAction, \
	WriteTopicAction
from .pipeline_graphic import PipelineGraphic, TopicGraphic, TopicRect
from .space import Space
from .topic import is_aggregation_topic, is_raw_topic, Topic, TopicKind, TopicType
from .topic_snapshot import TopicSnapshotFrequency, TopicSnapshotJobLock, TopicSnapshotJobLockId, \
	TopicSnapshotJobLockStatus, TopicSnapshotScheduler, TopicSnapshotSchedulerId
from .user import User, UserRole
from .user_group import UserGroup



from .catalog import Catalog, CatalogCriteria, CatalogId
from .dqc_pipelines import ask_dqc_pipelines
from .dqc_topics import ask_dqc_topics, MonitorRuleDetected
from .monitor_data import MonitorRuleLog, MonitorRuleLogCriteria
from .monitor_job_lock import MonitorJobLock, MonitorJobLockId
from .monitor_rule import MonitorRule, MonitorRuleCode, MonitorRuleCompareOperator, MonitorRuleGrade, MonitorRuleId, \
	MonitorRuleParameters, MonitorRuleSeverity, MonitorRuleStatisticalInterval
from .topic_profile import TopicProfile



from datetime import date, datetime
from enum import Enum
from typing import Optional, TypeVar

from pydantic import BaseModel

from watchmen_model.common import construct_parameter_joint, OptimisticLock, ParameterJoint, PipelineId, Storable, \
	TenantBasedTuple, TenantId, TopicId, UserId

TopicSnapshotSchedulerId = TypeVar('TopicSnapshotSchedulerId', bound=str)
TopicSnapshotJobLockId = TypeVar('TopicSnapshotJobLockId', bound=str)


class TopicSnapshotFrequency(str, Enum):
	DAILY = 'daily',
	WEEKLY = 'weekly',
	MONTHLY = 'monthly'


class TopicSnapshotScheduler(TenantBasedTuple, OptimisticLock, BaseModel):
	schedulerId: TopicSnapshotSchedulerId = None
	topicId: TopicId = None
	targetTopicName: str = None
	targetTopicId: TopicId = None
	pipelineId: PipelineId = None
	frequency: TopicSnapshotFrequency = TopicSnapshotFrequency.DAILY
	filter: Optional[ParameterJoint] = None
	# only for weekly
	weekday: Optional[str]
	# only for monthly
	day: Optional[str]
	hour: Optional[int] = None
	minute: Optional[int] = None
	enabled: bool = True

	def __setattr__(self, name, value):
		if name == 'filter':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


# noinspection DuplicatedCode
class TopicSnapshotJobLockStatus(str, Enum):
	READY = 'ready'
	FAILED = 'fail',
	SUCCESS = 'success'


class TopicSnapshotJobLock(Storable):
	lockId: TopicSnapshotJobLockId = None
	schedulerId: TopicSnapshotSchedulerId = None
	tenantId: TenantId = None
	frequency: TopicSnapshotFrequency = None
	processDate: date = None
	rowCount: int = 0
	status: TopicSnapshotJobLockStatus = None
	userId: UserId = None,
	createdAt: datetime = None



from enum import Enum
from typing import Any, Dict, TypeVar

from pydantic import BaseModel

from watchmen_model.admin import PipelineTriggerType
from watchmen_model.common import TenantId

PipelineTriggerTraceId = TypeVar('PipelineTriggerTraceId', bound=str)


class PipelineTriggerData(BaseModel):
	# topic name
	code: str = None
	# current data
	data: Dict[str, Any] = None
	triggerType: PipelineTriggerType = PipelineTriggerType.INSERT
	# pass tenant id when use super admin
	tenantId: TenantId = None
	# user given trace id, typically leave it as none
	traceId: PipelineTriggerTraceId = None


class PipelineTriggerDataWithPAT(PipelineTriggerData):
	pat: str


class PipelineTriggerResult(BaseModel):
	received: bool = True
	traceId: PipelineTriggerTraceId
	"""
	id of trigger data, 
	type must be str since length of value beyonds the limitation of serialization of javascript json number
	"""
	internalDataId: str


class TopicDataColumnNames(str, Enum):
	ID = 'id_',
	RAW_TOPIC_DATA = 'data_',
	AGGREGATE_ASSIST = 'aggregate_assist_',
	VERSION = 'version_',
	TENANT_ID = 'tenant_id_',
	INSERT_TIME = 'insert_time_',
	UPDATE_TIME = 'update_time_'



from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import Auditable, DashboardId, DataModel, GraphicRect, LastVisit, ReportId, UserBasedTuple
from watchmen_utilities import ArrayHelper
from .report import construct_funnels, ReportFunnel
from .utils import construct_rect


class DashboardReport(DataModel, BaseModel):
	reportId: ReportId = None
	funnels: List[ReportFunnel] = None
	rect: GraphicRect = None

	def __setattr__(self, name, value):
		if name == 'funnels':
			super().__setattr__(name, construct_funnels(value))
		elif name == 'rect':
			super().__setattr__(name, construct_rect(value))
		else:
			super().__setattr__(name, value)


class DashboardParagraph(DataModel, BaseModel):
	content: str = None
	rect: GraphicRect = None


def construct_report(report: Optional[Union[dict, DashboardReport]]) -> Optional[DashboardReport]:
	if report is None:
		return None
	elif isinstance(report, DashboardReport):
		return report
	else:
		return DashboardReport(**report)


def construct_reports(reports: List[Union[dict, DashboardReport]]) -> List[DashboardReport]:
	if reports is None:
		return []
	return ArrayHelper(reports).map(lambda x: construct_report(x)).to_list()


def construct_paragraph(paragraph: Optional[Union[dict, DashboardParagraph]]) -> Optional[DashboardParagraph]:
	if paragraph is None:
		return None
	elif isinstance(paragraph, DashboardParagraph):
		return paragraph
	else:
		return DashboardParagraph(**paragraph)


def construct_paragraphs(paragraphs: List[Union[dict, DashboardParagraph]]) -> List[DashboardParagraph]:
	if paragraphs is None:
		return []
	return ArrayHelper(paragraphs).map(lambda x: construct_paragraph(x)).to_list()


class Dashboard(UserBasedTuple, Auditable, LastVisit, BaseModel):
	dashboardId: DashboardId = None
	name: str = None
	reports: List[DashboardReport] = None
	paragraphs: List[DashboardParagraph] = None
	autoRefreshInterval: Optional[int] = None

	def __setattr__(self, name, value):
		if name == 'reports':
			super().__setattr__(name, construct_reports(value))
		elif name == 'paragraphs':
			super().__setattr__(name, construct_paragraphs(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import List

from pydantic import BaseModel

from watchmen_model.common import OptimisticLock, TenantBasedTuple, UserGroupId, UserId


class UserRole(str, Enum):
	CONSOLE = 'console',
	ADMIN = 'admin',
	# noinspection SpellCheckingInspection
	SUPER_ADMIN = 'superadmin'


class User(TenantBasedTuple, OptimisticLock, BaseModel):
	userId: UserId = None
	name: str = None
	nickName: str = None
	password: str = None
	isActive: bool = True
	groupIds: List[UserGroupId] = None
	role: UserRole = None



from typing import List

from watchmen_model.admin import Pipeline, Topic


# noinspection PyUnusedLocal
def ask_pipeline_monitor_pipelines(topics: List[Topic]) -> List[Pipeline]:
	# TODO define all pipeline monitor pipelines
	return []



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import BucketId, DataModel, FactorId
from watchmen_utilities import ArrayHelper, is_not_blank


class IndicatorCriteria(DataModel, BaseModel):
	factorId: FactorId = None


class IndicatorCriteriaOnBucket(IndicatorCriteria):
	"""
	fill when use predefined bucket
	"""
	bucketId: BucketId = None
	bucketSegmentName: str = None


class IndicatorCriteriaOperator(str, Enum):
	EQUALS = 'equals',
	NOT_EQUALS = 'not-equals',
	LESS = 'less',
	LESS_EQUALS = 'less-equals',
	MORE = 'more',
	MORE_EQUALS = 'more-equals',


class IndicatorCriteriaOnExpression(IndicatorCriteria):
	operator: IndicatorCriteriaOperator = IndicatorCriteriaOperator.EQUALS
	value: str = None


def construct_indicator_criteria(criteria: Optional[Union[dict, IndicatorCriteria]]) -> Optional[IndicatorCriteria]:
	if criteria is None:
		return None
	elif isinstance(criteria, IndicatorCriteria):
		return criteria
	else:
		bucket_id = criteria.get('bucketId')
		if is_not_blank(bucket_id):
			return IndicatorCriteriaOnBucket(**criteria)
		operator = criteria.get('operator')
		if is_not_blank(operator):
			return IndicatorCriteriaOnExpression(**criteria)
		else:
			return IndicatorCriteria(**criteria)


def construct_indicator_criteria_list(
		criteria_list: Optional[list] = None) -> Optional[List[IndicatorCriteria]]:
	if criteria_list is None:
		return None
	else:
		return ArrayHelper(criteria_list).map(lambda x: construct_indicator_criteria(x)).to_list()



from pydantic import BaseModel

from watchmen_model.common import DataModel
from .chart_enums import ChartBorderStyle, ChartFontStyle, ChartFontWeight
from .chart_types import ChartColor


class ChartFont(DataModel, BaseModel):
	family: str = None
	size: float = None
	color: ChartColor = None
	style: ChartFontStyle = None
	weight: ChartFontWeight = None


class ChartBorder(DataModel, BaseModel):
	color: ChartColor = None
	style: ChartBorderStyle = None
	width: float = None
	radius: float = None



from datetime import datetime
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import DataModel, GraphicPosition, GraphicRect, PipelineGraphicId, TopicId, UserBasedTuple
from watchmen_utilities import ArrayHelper


def construct_rect(rect: Optional[Union[dict, GraphicRect]]) -> Optional[GraphicRect]:
	if rect is None:
		return None
	elif isinstance(rect, GraphicRect):
		return rect
	else:
		return GraphicRect(**rect)


def construct_position(rect: Optional[Union[dict, GraphicPosition]]) -> Optional[GraphicPosition]:
	if rect is None:
		return None
	elif isinstance(rect, GraphicPosition):
		return rect
	else:
		return GraphicPosition(**rect)


class TopicRect(DataModel, BaseModel):
	coordinate: GraphicPosition = None
	frame: GraphicRect = None
	name: GraphicPosition = None

	def __setattr__(self, name, value):
		if name == 'frame':
			super().__setattr__(name, construct_rect(value))
		elif name == 'coordinate':
			super().__setattr__(name, construct_position(value))
		elif name == 'name':
			super().__setattr__(name, construct_position(value))
		else:
			super().__setattr__(name, value)


def construct_topic_rect(rect: Optional[Union[dict, TopicRect]]) -> Optional[TopicRect]:
	if rect is None:
		return None
	elif isinstance(rect, TopicRect):
		return rect
	else:
		return TopicRect(**rect)


class TopicGraphic(DataModel, BaseModel):
	topicId: TopicId = None
	rect: TopicRect = None

	def __setattr__(self, name, value):
		if name == 'rect':
			super().__setattr__(name, construct_topic_rect(value))
		else:
			super().__setattr__(name, value)


# noinspection DuplicatedCode
def construct_topic(topic: Optional[Union[dict, TopicGraphic]]) -> Optional[TopicGraphic]:
	if topic is None:
		return None
	elif isinstance(topic, TopicGraphic):
		return topic
	else:
		return TopicGraphic(**topic)


def construct_topics(topics: List[Union[dict, TopicGraphic]]) -> List[TopicGraphic]:
	if topics is None:
		return []
	return ArrayHelper(topics).map(lambda x: construct_topic(x)).to_list()


class PipelineGraphic(UserBasedTuple, BaseModel):
	pipelineGraphId: PipelineGraphicId = None
	name: str = None
	topics: List[TopicGraphic] = []
	createdAt: datetime = None
	lastModifiedAt: datetime = None

	def __setattr__(self, name, value):
		if name == 'topics':
			super().__setattr__(name, construct_topics(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import Any, List, Literal, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import BucketId, DataModel, EnumId, OptimisticLock, TenantBasedTuple
from watchmen_utilities import ArrayHelper
from .measure_method import MeasureMethod


class BucketType(str, Enum):
	VALUE = 'value'
	VALUE_MEASURE = 'value-measure'
	CATEGORY_MEASURE = 'category-measure'
	ENUM_MEASURE = 'enum-measure'


class RangeBucketValueIncluding(str, Enum):
	INCLUDE_MIN = "include-min"
	INCLUDE_MAX = "include-max"


class BucketSegment(DataModel, BaseModel):
	name: str = None
	value: Any = None


def construct_bucket_segment(
		segment: Optional[Union[dict, BucketSegment]], bucket_type: BucketType) -> Optional[BucketSegment]:
	if segment is None:
		return None
	elif isinstance(segment, BucketSegment):
		return segment
	elif bucket_type == BucketType.VALUE or bucket_type == BucketType.VALUE_MEASURE:
		return NumericValueSegment(**segment)
	elif bucket_type == BucketType.ENUM_MEASURE or bucket_type == BucketType.CATEGORY_MEASURE:
		return CategorySegment(**segment)
	else:
		raise Exception(f'Bucket type[{bucket_type}] cannot be recognized.')


def construct_bucket_segments(
		subjects: List[Union[dict, BucketSegment]], bucket_type: BucketType) -> List[BucketSegment]:
	if subjects is None:
		return []
	return ArrayHelper(subjects).map(lambda x: construct_bucket_segment(x, bucket_type)).to_list()


class Bucket(TenantBasedTuple, OptimisticLock, BaseModel):
	bucketId: BucketId = None
	name: str = None
	type: BucketType = None
	segments: List[BucketSegment] = None
	description: str = None

	def __setattr__(self, name, value):
		if name == 'segments':
			if self.type is not None:
				super().__setattr__(name, construct_bucket_segments(value, self.type))
			else:
				super().__setattr__(name, value)
		elif name == 'type':
			if self.segments is not None:
				super().__setattr__(name, construct_bucket_segments(self.segments, value))
			else:
				super().__setattr__(name, value)
		else:
			super().__setattr__(name, value)


class NumericSegmentValue(DataModel, BaseModel):
	min: str = None
	max: str = None


def construct_numeric_segment_value(
		segment_value: Optional[Union[dict, NumericSegmentValue]]) -> Optional[NumericSegmentValue]:
	if segment_value is None:
		return None
	elif isinstance(segment_value, NumericSegmentValue):
		return segment_value
	else:
		return NumericSegmentValue(**segment_value)


class NumericValueSegment(BucketSegment):
	value: NumericSegmentValue

	def __setattr__(self, name, value):
		if name == 'value':
			super().__setattr__(name, construct_numeric_segment_value(value))
		else:
			super().__setattr__(name, value)


class NumericSegmentsHolder(Bucket):
	include: RangeBucketValueIncluding = None
	segments: List[NumericValueSegment]


class NumericValueBucket(NumericSegmentsHolder):
	type: BucketType.VALUE = BucketType.VALUE


class MeasureBucket(Bucket):
	measure: MeasureMethod = None


class NumericValueMeasureBucket(MeasureBucket, NumericSegmentsHolder):
	type: BucketType.VALUE_MEASURE = BucketType.VALUE_MEASURE
	measure: Union[
		MeasureMethod.FLOOR, MeasureMethod.RESIDENTIAL_AREA, MeasureMethod.AGE, MeasureMethod.BIZ_SCALE
	] = None


OtherCategorySegmentValue = Literal['&others']

CategorySegmentValue = List[Union[str, OtherCategorySegmentValue]]


class CategorySegment(BucketSegment):
	value: CategorySegmentValue


class CategorySegmentsHolder(Bucket):
	segments: List[CategorySegment]


class CategoryMeasureBucket(CategorySegmentsHolder, MeasureBucket):
	type: BucketType.CATEGORY_MEASURE = BucketType.CATEGORY_MEASURE
	measure: Union[
		MeasureMethod.CONTINENT, MeasureMethod.REGION, MeasureMethod.COUNTRY,
		MeasureMethod.PROVINCE, MeasureMethod.CITY, MeasureMethod.DISTRICT,
		MeasureMethod.RESIDENCE_TYPE,
		MeasureMethod.GENDER, MeasureMethod.OCCUPATION, MeasureMethod.RELIGION, MeasureMethod.NATIONALITY,
		MeasureMethod.BIZ_TRADE,
		MeasureMethod.BOOLEAN
	]


class EnumMeasureBucket(CategorySegmentsHolder, MeasureBucket):
	type: BucketType.ENUM_MEASURE = BucketType.ENUM_MEASURE
	measure: MeasureMethod.ENUM = MeasureMethod.ENUM
	enumId: EnumId



from enum import Enum


class MeasureMethod(str, Enum):
	# address related
	CONTINENT = 'continent',
	REGION = 'region',
	COUNTRY = 'country',
	PROVINCE = 'province',
	CITY = 'city',
	DISTRICT = 'district',
	FLOOR = 'floor',
	RESIDENCE_TYPE = 'residence-type',
	RESIDENTIAL_AREA = 'residential-area',

	# time related
	YEAR = 'year',
	HALF_YEAR = 'half-year',
	QUARTER = 'quarter',
	MONTH = 'month',
	HALF_MONTH = 'half-month',
	TEN_DAYS = 'ten-days',
	WEEK_OF_YEAR = 'week-of-year',
	WEEK_OF_MONTH = 'week-of-month',
	HALF_WEEK = 'half-week',
	DAY_OF_MONTH = 'day-of-month',
	DAY_OF_WEEK = 'day-of-week',
	DAY_KIND = 'day-kind',
	HOUR = 'hour',
	HOUR_KIND = 'hour-kind',
	AM_PM = 'am-pm',

	# individual related
	GENDER = 'gender',
	OCCUPATION = 'occupation',
	AGE = 'age',
	RELIGION = 'religion',
	NATIONALITY = 'nationality',

	# organization related
	BIZ_TRADE = 'biz-trade',
	BIZ_SCALE = 'biz-scale',

	# boolean
	BOOLEAN = 'boolean',

	# enumeration
	ENUM = 'enum',



from watchmen_model.common import construct_parameter_joint
from .pipeline_action import DeleteTopicActionType, FindBy, ToTopic


class DeleteTopicAction(ToTopic, FindBy):
	type: DeleteTopicActionType = None

	def __setattr__(self, name, value):
		if name == 'by':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


class DeleteRowAction(DeleteTopicAction):
	type: DeleteTopicActionType = DeleteTopicActionType.DELETE_ROW


class DeleteRowsAction(DeleteTopicAction):
	type: DeleteTopicActionType = DeleteTopicActionType.DELETE_ROWS



from typing import Any, Dict, List, Union

from pydantic import BaseModel

from .model import DataModel


class Pageable(DataModel, BaseModel):
	pageNumber: int = None
	pageSize: int = None


PageDataCell = Any
"""
data row with name
"""
PageDataRow = Union[Dict[str, PageDataCell], DataModel]
PageDataSet = Union[List[Dict[str, PageDataCell]], List[DataModel]]


class DataPage(Pageable):
	"""
	PageDataSet
	"""
	data: list = []
	itemCount: int = None
	pageCount: int = None



from enum import Enum

from pydantic import BaseModel

from watchmen_model.common import DataModel


class ChartTruncationType(str, Enum):
	NONE = 'none',
	TOP = 'top',
	BOTTOM = 'bottom'


class ChartTruncation(DataModel, BaseModel):
	type: ChartTruncationType = ChartTruncationType.NONE
	count: int = 20


class ChartTruncationHolder(DataModel, BaseModel):
	truncation: ChartTruncation = None



from datetime import date, datetime

from pydantic import BaseConfig

from .storable import Auditable, Storable
from .tuple_ids import TenantId, UserId

"""
Super model of tuple, which 
1. contains 4 audit properties. There is no timezone for datetime,
2. align the json serialization/deserialization format
"""


class Tuple(Auditable):
	class Config(BaseConfig):
		json_encoders = {
			datetime: lambda dt: dt.isoformat(),
			date: lambda dt: dt.isoformat()
		}


class TenantBasedTuple(Tuple):
	tenantId: TenantId = None


class UserBasedTuple(Storable):
	"""
	no audit columns
	"""
	tenantId: TenantId = None
	userId: UserId = None

	class Config(BaseConfig):
		json_encoders = {
			datetime: lambda dt: dt.isoformat(),
			date: lambda dt: dt.isoformat()
		}



from .connected_space import ConnectedSpace
from .connected_space_graphic import ConnectedSpaceGraphic, SubjectGraphic, TopicGraphic
from .dashboard import Dashboard, DashboardParagraph, DashboardReport
from .data_result_set import SubjectDataResultSet
from .report import Report, ReportDimension, ReportFunnel, ReportFunnelType, ReportIndicator, ReportIndicatorArithmetic
from .subject import Subject, SubjectDataset, SubjectDatasetColumn, SubjectDatasetCriteria, \
	SubjectDatasetCriteriaIndicator, SubjectDatasetCriteriaIndicatorArithmetic, SubjectDatasetJoin, SubjectJoinType



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.chart import BarChart, Chart, ChartType, CountChart, CustomizedChart, DoughnutChart, LineChart, \
	MapChart, NightingaleChart, PieChart, ScatterChart, SunburstChart, TreeChart, TreemapChart
from watchmen_model.common import Auditable, ConnectedSpaceId, construct_parameter_joint, DataModel, DataResultSet, \
	GraphicRect, LastVisit, ParameterJoint, ReportFunnelId, ReportId, SubjectDatasetColumnId, SubjectId, UserBasedTuple
from watchmen_utilities import ArrayHelper
from .utils import construct_rect


class ReportIndicatorArithmetic(str, Enum):
	NONE = 'none'
	COUNT = 'count'
	SUMMARY = 'sum'
	AVERAGE = 'avg'
	MAXIMUM = 'max'
	MINIMUM = 'min'


class ReportIndicator(DataModel, BaseModel):
	columnId: SubjectDatasetColumnId = None
	name: str = None
	arithmetic: ReportIndicatorArithmetic = ReportIndicatorArithmetic.NONE


class ReportDimension(DataModel, BaseModel):
	columnId: SubjectDatasetColumnId = None
	name: str = None


class ReportFunnelType(str, Enum):
	NUMERIC = 'numeric',
	DATE = 'date',
	YEAR = 'year',
	HALF_YEAR = 'half-year',
	QUARTER = 'quarter',
	MONTH = 'month',
	HALF_MONTH = 'half-month',
	TEN_DAYS = 'ten-days',
	WEEK_OF_MONTH = 'week-of-month',
	HALF_WEEK = 'half-week',
	DAY_KIND = 'day-kind',
	DAY_OF_WEEK = 'day-of-week',
	HOUR = 'hour',
	HOUR_KIND = 'hour-kind',
	AM_PM = 'am-pm',
	ENUM = 'enum'


class ReportFunnel(DataModel, BaseModel):
	funnelId: ReportFunnelId = None
	columnId: SubjectDatasetColumnId = None
	type: ReportFunnelType = None
	range: bool = False
	enabled: bool = False
	values: List[Union[str, None]] = None


def construct_chart(chart: Optional[Union[dict, Chart]]) -> Optional[Chart]:
	if chart is None:
		return None
	elif isinstance(chart, Chart):
		return chart
	else:
		chart_type = chart.get('type')
		if chart_type == ChartType.COUNT:
			return CountChart(**chart)
		elif chart_type == ChartType.BAR:
			return BarChart(**chart)
		elif chart_type == ChartType.LINE:
			return LineChart(**chart)
		elif chart_type == ChartType.SCATTER:
			return ScatterChart(**chart)
		elif chart_type == ChartType.PIE:
			return PieChart(**chart)
		elif chart_type == ChartType.DOUGHNUT:
			return DoughnutChart(**chart)
		elif chart_type == ChartType.NIGHTINGALE:
			return NightingaleChart(**chart)
		elif chart_type == ChartType.SUNBURST:
			return SunburstChart(**chart)
		elif chart_type == ChartType.TREE:
			return TreeChart(**chart)
		elif chart_type == ChartType.TREEMAP:
			return TreemapChart(**chart)
		elif chart_type == ChartType.MAP:
			return MapChart(**chart)
		elif chart_type == ChartType.CUSTOMIZED:
			return CustomizedChart(**chart)
		else:
			raise Exception(f'Chart type[{chart_type}] cannot be recognized.')


def construct_funnel(funnel: Optional[Union[dict, ReportFunnel]]) -> Optional[ReportFunnel]:
	if funnel is None:
		return None
	elif isinstance(funnel, ReportFunnel):
		return funnel
	else:
		return ReportFunnel(**funnel)


def construct_funnels(funnels: List[Union[dict, ReportFunnel]]) -> List[ReportFunnel]:
	if funnels is None:
		return []
	return ArrayHelper(funnels).map(lambda x: construct_funnel(x)).to_list()


def construct_indicator(indicator: Optional[Union[dict, ReportIndicator]]) -> Optional[ReportIndicator]:
	if indicator is None:
		return None
	elif isinstance(indicator, ReportIndicator):
		return indicator
	else:
		return ReportIndicator(**indicator)


def construct_indicators(indicators: List[Union[dict, ReportIndicator]]) -> List[ReportIndicator]:
	if indicators is None:
		return []
	return ArrayHelper(indicators).map(lambda x: construct_indicator(x)).to_list()


def construct_dimension(dimension: Optional[Union[dict, ReportDimension]]) -> Optional[ReportDimension]:
	if dimension is None:
		return None
	elif isinstance(dimension, ReportDimension):
		return dimension
	else:
		return ReportDimension(**dimension)


def construct_dimensions(dimensions: List[Union[dict, ReportDimension]]) -> List[ReportDimension]:
	if dimensions is None:
		return []
	return ArrayHelper(dimensions).map(lambda x: construct_dimension(x)).to_list()


# noinspection PyRedundantParentheses
class AvoidFastApiError:
	filters: ParameterJoint = None


class Report(UserBasedTuple, Auditable, LastVisit, AvoidFastApiError, BaseModel):
	reportId: ReportId = None
	name: str = None
	subjectId: SubjectId = None
	connectId: ConnectedSpaceId = None
	funnels: List[ReportFunnel] = None
	indicators: List[ReportIndicator] = None
	dimensions: List[ReportDimension] = None
	description: str = None
	rect: GraphicRect = None
	chart: Chart = None
	simulating: bool = False
	simulateData: DataResultSet = None
	# base64
	simulateThumbnail: str = None

	def __setattr__(self, name, value):
		if name == 'filters':
			super().__setattr__(name, construct_parameter_joint(value))
		elif name == 'chart':
			super().__setattr__(name, construct_chart(value))
		elif name == 'rect':
			super().__setattr__(name, construct_rect(value))
		elif name == 'funnels':
			super().__setattr__(name, construct_funnels(value))
		elif name == 'indicators':
			super().__setattr__(name, construct_indicators(value))
		elif name == 'dimensions':
			super().__setattr__(name, construct_dimensions(value))
		else:
			super().__setattr__(name, value)



from enum import Enum

from pydantic import BaseModel

from watchmen_model.common import ExternalWriterId, OptimisticLock, TenantBasedTuple


class ExternalWriterType(str, Enum):
	STANDARD_WRITER = 'standard-writer',
	ELASTIC_SEARCH_WRITER = 'elastic-search-writer'


class ExternalWriter(TenantBasedTuple, OptimisticLock, BaseModel):
	writerId: ExternalWriterId = None
	writerCode: str = None
	type: ExternalWriterType = ExternalWriterType.STANDARD_WRITER
	# personal access token
	pat: str = None
	url: str = None



from enum import Enum


class PredefinedChartColorSeries(str, Enum):
	REGULAR = 'regular',
	DARK = 'dark',
	LIGHT = 'light'


class ChartBorderStyle(str, Enum):
	NONE = 'none',
	SOLID = 'solid',
	DOTTED = 'dotted',
	DASHED = 'dashed'


class ChartFontStyle(str, Enum):
	NORMAL = 'normal',
	ITALIC = 'italic'


class ChartFontWeight(str, Enum):
	W100 = '100',
	W200 = '200',
	W300 = '300',
	W400 = '400',
	W500 = '500',
	W600 = '600',
	W700 = '700',
	W800 = '800',
	W900 = '900'



from typing import List, Optional, TypeVar

from pydantic import BaseModel

from watchmen_model.common import OptimisticLock, TenantBasedTuple, TopicId, UserId

CatalogId = TypeVar('CatalogId', bound=str)


class Catalog(TenantBasedTuple, OptimisticLock, BaseModel):
	catalogId: CatalogId = None
	name: str = None
	topicIds: List[TopicId] = []
	techOwnerId: UserId = None
	bizOwnerId: UserId = None
	tags: List[str] = []
	description: str = None


class CatalogCriteria(BaseModel):
	name: Optional[str] = None
	topicId: Optional[TopicId] = None
	techOwnerId: Optional[UserId] = None
	bizOwnerId: Optional[UserId] = None



from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import DataModel
from .chart_basic_style import ChartBorder, ChartBorderStyle, ChartFont
from .chart_settings import ChartSettings
from .chart_types import ChartColor, ChartType


class Chart(DataModel, BaseModel):
	type: ChartType = ChartType.COUNT
	settings: ChartSettings = None

	def __setattr__(self, name, value):
		if name == 'settings':
			if self.type is not None:
				super().__setattr__(name, construct_settings(value, self.type))
			else:
				super().__setattr__(name, value)
		elif name == 'type':
			if self.settings is not None:
				super().__setattr__(name, construct_settings(self.settings, value))
			else:
				super().__setattr__(name, value)
		else:
			super().__setattr__(name, value)


class EChartsBorderHolder(DataModel):
	border: ChartBorder = None


class EChartsBorderOmitRadius(DataModel):
	color: ChartColor = None
	style: ChartBorderStyle = None
	width: float = None


class EChartsBorderHolderNoRadius(DataModel):
	border: EChartsBorderOmitRadius = None


class EChartsFontHolder(DataModel):
	font: ChartFont = None


class EChartsPosition(DataModel):
	top: float = None
	right: float = None
	left: float = None
	bottom: float = None


class EChartsPositionHolder(DataModel):
	position: EChartsPosition = None


class EChartsHorizontalAlignment(str, Enum):
	AUTO = 'auto',
	LEFT = 'left',
	RIGHT = 'right',
	CENTER = 'center'


class EChartsVerticalAlignment(str, Enum):
	AUTO = 'auto',
	TOP = 'top',
	BOTTOM = 'bottom',
	MIDDLE = 'middle'


class EChartsAlignmentHolder(DataModel):
	horizontalAlign: EChartsHorizontalAlignment = None
	verticalAlign: EChartsVerticalAlignment = None


class EChartsTitleText(EChartsFontHolder, BaseModel):
	text: str = None


class EChartsTitle(EChartsBorderHolder, EChartsPositionHolder, EChartsAlignmentHolder, BaseModel):
	text: EChartsTitleText = None
	subtext: EChartsTitleText = None
	backgroundColor: ChartColor = None
	padding: float = None
	itemGap: float = None


class EChartsTitleHolder(DataModel, BaseModel):
	title: EChartsTitle = None


class EChartsLegendOrient(str, Enum):
	HORIZONTAL = 'horizontal',
	VERTICAL = 'vertical'


class EChartsLegend(EChartsBorderHolder, EChartsPositionHolder, EChartsFontHolder):
	show: bool = None
	orient: EChartsLegendOrient = None
	backgroundColor: ChartColor = None
	padding: float = None


class EChartsLegendHolder(DataModel):
	legend: EChartsLegend = None


class EChartsGrid(EChartsBorderHolderNoRadius, EChartsPositionHolder):
	show: bool = None
	containLabel: bool = None
	backgroundColor: ChartColor = None


class EChartsGridPositionOnly(EChartsPositionHolder, BaseModel):
	pass


class EChartsGridHolder(DataModel):
	grid: EChartsGrid = None


class EChartsAxisSplitLineStyle(str, Enum):
	SOLID = 'solid',
	DASHED = 'dashed',
	DOTTED = 'dotted'


class EChartsAxisSplitLine(DataModel):
	show: bool = None
	color: ChartColor = None
	width: float = None
	style: EChartsAxisSplitLineStyle = None


class EChartsAxisSplitLineHolder(DataModel):
	splitLine: EChartsAxisSplitLine = None


class EChartsAxisMinorSplitLineHolder(DataModel):
	minorSplitLine: EChartsAxisSplitLine = None


class EChartsXAxisPosition(str, Enum):
	TOP = 'top',
	BOTTOM = 'bottom'


class EChartsXAxisType(str, Enum):
	VALUE = 'value',
	CATEGORY = 'category',
	TIME = 'time'


class EChartsXAxisNameLocation(str, Enum):
	START = 'start',
	CENTER = 'center',
	END = 'end'


# noinspection DuplicatedCode
class EChartsXAxisName(EChartsFontHolder, EChartsBorderHolder, EChartsAlignmentHolder):
	text: str = None
	location: EChartsXAxisNameLocation = None
	backgroundColor: ChartColor = None
	gap: float = None
	rotate: float = None
	padding: float = None


class EChartsXAxisLabel(EChartsFontHolder, EChartsBorderHolder, EChartsAlignmentHolder):
	show: bool = None
	inside: bool = None
	backgroundColor: ChartColor = None
	gap: float = None
	rotate: float = None
	padding: float = None


class EChartsXAxis(EChartsAxisSplitLineHolder, EChartsAxisMinorSplitLineHolder):
	show: bool = None
	position: EChartsXAxisPosition = None
	type: EChartsXAxisType = None
	name: EChartsXAxisName = None
	label: EChartsXAxisLabel = None
	autoMin: bool = None
	min: float = None
	autoMax: bool = None
	max: float = None


class EChartsXAxisHolder(DataModel):
	# noinspection SpellCheckingInspection
	xaxis: EChartsXAxis = None


class EChartsYAxisPosition(str, Enum):
	LEFT = 'left',
	RIGHT = 'right'


class EChartsYAxisType(str, Enum):
	VALUE = 'value',
	CATEGORY = 'category',
	TIME = 'time'


class EChartsYAxisNameLocation(str, Enum):
	START = 'start',
	MIDDLE = 'middle',
	END = 'end'


# noinspection DuplicatedCode
class EChartsYAxisName(EChartsFontHolder, EChartsBorderHolder, EChartsAlignmentHolder):
	text: str = None
	location: EChartsYAxisNameLocation = None
	backgroundColor: ChartColor = None
	gap: float = None
	rotate: float = None
	padding: float = None


class EChartsYAxisLabel(EChartsFontHolder, EChartsBorderHolder, EChartsAlignmentHolder):
	show: bool = None
	inside: bool = None
	backgroundColor: ChartColor = None
	gap: float = None
	rotate: float = None
	padding: float = None


class EChartsYAxis(EChartsAxisSplitLineHolder, EChartsAxisMinorSplitLineHolder):
	show: bool = None
	position: EChartsYAxisPosition = None
	type: EChartsYAxisType = None
	name: EChartsYAxisName = None
	label: EChartsYAxisLabel = None
	autoMin: bool = None
	min: float = None
	autoMax: bool = None
	max: float = None


class EChartsYAxisHolder(DataModel):
	yaxis: EChartsYAxis = None


class EChartsToolboxOrient(str, Enum):
	HORIZONTAL = 'horizontal',
	VERTICAL = 'vertical'


class EChartsToolbox(EChartsPositionHolder):
	show: bool = None
	orient: EChartsToolboxOrient = None


class EChartsToolboxHolder(DataModel):
	toolbox: EChartsToolbox = None


EchartsScriptsVars = Dict[str, str]


class EchartsScriptHolder(DataModel):
	script: str = None
	# noinspection SpellCheckingInspection
	scriptVarsDefs: str = None
	scriptVars: EchartsScriptsVars = None


class ItemType(str, Enum):
	SECTION = 'section',
	NUMBER = 'number',
	PERCENTAGE = 'percentage',
	BOOLEAN = 'boolean',
	TEXT = 'text',
	COLOR = 'color',
	DROPDOWN = 'dropdown'


class DefItem(DataModel):
	type: ItemType = None
	label: str = None


class SectionItem(DefItem):
	type: ItemType.SECTION = ItemType.SECTION


class InputItem(DefItem):
	key: str = None


class NumberItem(InputItem):
	type: ItemType.NUMBER = ItemType.NUMBER
	placeholder: str = None
	unit: str = None
	defaultValue: float = None


class PercentageItem(InputItem):
	type: ItemType.PERCENTAGE = ItemType.PERCENTAGE
	placeholder: str = None
	defaultValue: float = None


class BooleanItem(InputItem):
	type: ItemType.BOOLEAN = ItemType.BOOLEAN
	defaultValue: bool = None


class TextItem(InputItem):
	type: ItemType.TEXT = ItemType.TEXT
	placeholder: str = None
	defaultValue: str = None


class ColorItem(InputItem):
	type: ItemType.COLOR = ItemType.COLOR
	defaultValue: ChartColor = None


class DropdownItemOption(DataModel):
	value: Union[str, int, float, bool] = None
	label: str = None


class DropdownItem(InputItem):
	type: ItemType.DROPDOWN = ItemType.DROPDOWN
	placeholder: str = None
	options: List[DropdownItemOption] = []
	defaultValue: Any = None


class EChartsSettings(ChartSettings, EChartsTitleHolder):
	pass


class CountChartSettingsText(DataModel, BaseModel):
	font: ChartFont = None
	formatUseGrouping: bool = None


class CountChartSettings(EChartsSettings):
	countText: CountChartSettingsText = None


class CountChart(Chart):
	type: ChartType = ChartType.COUNT
	settings: CountChartSettings = None


class BarLabelPosition(str, Enum):
	TOP = 'top',
	LEFT = 'left',
	RIGHT = 'right',
	BOTTOM = 'bottom',
	INSIDE = 'inside',
	INSIDE_LEFT = 'insideLeft',
	INSIDE_RIGHT = 'insideRight',
	INSIDE_TOP = 'insideTop',
	INSIDE_BOTTOM = 'insideBottom',
	INSIDE_TOP_LEFT = 'insideTopLeft',
	INSIDE_BOTTOM_LEFT = 'insideBottomLeft',
	INSIDE_TOP_RIGHT = 'insideTopRight',
	INSIDE_BOTTOM_RIGHT = 'insideBottomRight'


class BarChartSettingsLabel(EChartsBorderHolder, EChartsFontHolder, EChartsAlignmentHolder, BaseModel):
	show: bool = None
	backgroundColor: ChartColor = None
	position: BarLabelPosition = None
	rotate: float = None
	gap: float = None
	padding: float = None
	formatUseGrouping: bool = None
	formatUsePercentage: bool = None
	valueAsPercentage: bool = None
	fractionDigits: int = None


class BarChartSettingsSeries(DataModel, BaseModel):
	transformAxis: bool = None


class BarChartSettings(EChartsSettings, EChartsLegendHolder, EChartsGridHolder, EChartsXAxisHolder, EChartsYAxisHolder):
	series: BarChartSettingsSeries = None
	label: BarChartSettingsLabel = None
	decal: bool = None


class BarChart(Chart):
	type: ChartType = ChartType.BAR
	settings: BarChartSettings = None


class LineChartSettingsSeries(BarChartSettingsSeries):
	smooth: bool = None


class LineChartSettings(
	EChartsSettings, EChartsLegendHolder, EChartsGridHolder, EChartsXAxisHolder, EChartsYAxisHolder):
	series: LineChartSettingsSeries = None
	label: BarChartSettingsLabel = None


class LineChart(Chart):
	type: ChartType = ChartType.LINE
	settings: LineChartSettings = None


class ScatterChartSettings(
	EChartsSettings, EChartsLegendHolder, EChartsGridHolder, EChartsXAxisHolder, EChartsYAxisHolder):
	pass


class ScatterChart(Chart):
	type: ChartType = ChartType.SCATTER
	settings: ScatterChartSettings = None


class PieRoseType(str, Enum):
	NONE = 'none',
	RADIUS = 'radius',
	AREA = 'area'


class PieLabelPosition(str, Enum):
	INSIDE = 'inside',
	OUTSIDE = 'outside',
	CENTER = 'center'


class PieLabelAlignTo(str, Enum):
	NONE = 'none',
	LABEL_LINE = 'labelLine',
	EDGE = 'edge'


# noinspection DuplicatedCode
class PieChartSettingsLabel(EChartsBorderHolder, EChartsFontHolder, EChartsAlignmentHolder, BaseModel):
	show: bool = None
	backgroundColor: ChartColor = None
	position: PieLabelPosition = None
	alignTo: PieLabelAlignTo = None
	rotate: float = None
	gap: float = None
	padding: float = None
	formatUseGrouping: bool = None
	formatUsePercentage: bool = None
	valueAsPercentage: bool = None
	fractionDigits: int = None


class PieChartSettingsSeries(EChartsBorderHolder, BaseModel):
	centerX: float = None
	centerY: float = None
	insideRadius: float = None
	outsideRadius: float = None
	roseType: PieRoseType = None
	showPercentage: bool = None


class PieChartSettings(EChartsSettings, EChartsLegendHolder):
	series: PieChartSettingsSeries = None
	grid: EChartsGridPositionOnly = None
	label: PieChartSettingsLabel = None
	decal: bool = None


class PieChart(Chart):
	type: ChartType = ChartType.PIE
	settings: PieChartSettings = None


class DoughnutChartSettings(PieChartSettings):
	pass


class DoughnutChart(Chart):
	type: ChartType = ChartType.DOUGHNUT
	settings: DoughnutChartSettings = None


class NightingaleChartSettings(PieChartSettings):
	pass


class NightingaleChart(Chart):
	type: ChartType = ChartType.NIGHTINGALE
	settings: NightingaleChartSettings = None


class SunburstChartSettingsSeries(EChartsBorderHolder, BaseModel):
	centerX: float = None
	centerY: float = None
	insideRadius: float = None
	outsideRadius: float = None


class SunburstChartSettings(PieChartSettings):
	series: SunburstChartSettingsSeries = None


class SunburstChart(Chart):
	type: ChartType = ChartType.SUNBURST
	settings: SunburstChartSettings = None


class TreeLayout(str, Enum):
	ORTHOGONAL = 'orthogonal',
	RADIAL = 'radial'


class TreeOrient(str, Enum):
	LEFT_RIGHT = 'LR',
	RIGHT_LEFT = 'RL',
	TOP_BOTTOM = 'TB',
	BOTTOM_TOP = 'BT'


class TreeChartSettingsSeries(DataModel, BaseModel):
	layout: TreeLayout = None
	orient: TreeOrient = None
	roam: bool = None


class TreeChartSettings(EChartsSettings):
	series: TreeChartSettingsSeries = None
	grid: EChartsGridPositionOnly = None


class TreeChart(Chart):
	type: ChartType = ChartType.TREE
	settings: TreeChartSettings = None


class TreemapChartSettingsSeries(DataModel, BaseModel):
	roam: bool = None


class TreemapChartSettings(EChartsSettings):
	series: TreemapChartSettingsSeries = None
	grid: EChartsGridPositionOnly = None


class TreemapChart(Chart):
	type: ChartType = ChartType.TREEMAP
	settings: TreemapChartSettings = None


class MapChartRegion(str, Enum):
	CHINA_L1 = 'china-l1',
	CYPRUS_L1 = 'cyprus-l1',
	JAPAN_L1 = 'japan-l1',
	SINGAPORE_L1 = 'singapore-l1',
	USA_L1 = 'usa-l1'


class MapChartSettingsSeries(DataModel, BaseModel):
	region: MapChartRegion = None


class MapChartSettings(EChartsSettings):
	series: MapChartSettingsSeries = None
	grid: EChartsGridPositionOnly = None


class MapChart(Chart):
	type: ChartType = ChartType.MAP
	settings: MapChartSettings = None


class CustomizedChartSettings(EChartsSettings, EchartsScriptHolder):
	pass


class CustomizedChart(Chart):
	type: ChartType = ChartType.CUSTOMIZED
	settings: CustomizedChartSettings = None


def construct_settings(
		settings: Optional[Union[dict, ChartSettings]], chart_type: ChartType) -> Optional[ChartSettings]:
	if settings is None:
		return None
	elif isinstance(settings, ChartSettings):
		return settings
	elif chart_type == ChartType.COUNT:
		return CountChartSettings(**settings)
	elif chart_type == ChartType.BAR:
		return BarChartSettings(**settings)
	elif chart_type == ChartType.LINE:
		return LineChartSettings(**settings)
	elif chart_type == ChartType.SCATTER:
		return ScatterChartSettings(**settings)
	elif chart_type == ChartType.PIE:
		return PieChartSettings(**settings)
	elif chart_type == ChartType.DOUGHNUT:
		return DoughnutChartSettings(**settings)
	elif chart_type == ChartType.NIGHTINGALE:
		return NightingaleChartSettings(**settings)
	elif chart_type == ChartType.SUNBURST:
		return SunburstChartSettings(**settings)
	elif chart_type == ChartType.TREE:
		return TreeChartSettings(**settings)
	elif chart_type == ChartType.TREEMAP:
		return TreemapChartSettings(**settings)
	elif chart_type == ChartType.MAP:
		return MapChartSettings(**settings)
	elif chart_type == ChartType.CUSTOMIZED:
		return CustomizedChartSettings(**settings)
	else:
		raise Exception(f'Chart type[{chart_type}] cannot be recognized.')



from datetime import date, datetime
from enum import Enum
from typing import TypeVar

from watchmen_model.common import Storable, TenantId, TopicId, UserId
from .monitor_rule import MonitorRuleStatisticalInterval

MonitorJobLockId = TypeVar('MonitorJobLockId', bound=str)


# noinspection DuplicatedCode
class MonitorJobLockStatus(str, Enum):
	READY = 'ready'
	FAILED = 'fail',
	SUCCESS = 'success'


class MonitorJobLock(Storable):
	lockId: MonitorJobLockId = None
	tenantId: TenantId = None
	topicId: TopicId = None
	frequency: MonitorRuleStatisticalInterval = None
	processDate: date = None
	status: MonitorJobLockStatus = None
	userId: UserId = None,
	createdAt: datetime = None



from __future__ import annotations

from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_utilities import ArrayHelper
from .model import DataModel
from .tuple_ids import FactorId, TopicId


class ParameterKind(str, Enum):
	TOPIC = 'topic',
	CONSTANT = 'constant',
	COMPUTED = 'computed'


class AvoidFastApiError:
	on: Optional[ParameterJoint] = None


class Parameter(DataModel, AvoidFastApiError, BaseModel):
	kind: ParameterKind = None
	conditional: bool = False

	def __setattr__(self, name, value):
		if name == 'on':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


class TopicFactorParameter(Parameter):
	kind: ParameterKind.TOPIC = ParameterKind.TOPIC
	topicId: TopicId = None
	factorId: FactorId = None


class VariablePredefineFunctions(str, Enum):
	NEXT_SEQ = '&nextSeq',
	COUNT = '&count',
	LENGTH = '&length',
	SUM = '&sum',
	FROM_PREVIOUS_TRIGGER_DATA = '&old',

	DAY_DIFF = '&dayDiff',
	MONTH_DIFF = '&monthDiff',
	YEAR_DIFF = '&yearDiff'
	DATE_FORMAT = '&fmtDate'
	NOW = '&now'


class ConstantParameter(Parameter):
	kind: ParameterKind.CONSTANT = ParameterKind.CONSTANT
	value: str = None


class ParameterComputeType(str, Enum):
	NONE = 'none',
	ADD = 'add',
	SUBTRACT = 'subtract',
	MULTIPLY = 'multiply',
	DIVIDE = 'divide',
	MODULUS = 'modulus',
	YEAR_OF = 'year-of',
	HALF_YEAR_OF = 'half-year-of',
	QUARTER_OF = 'quarter-of',
	MONTH_OF = 'month-of',
	WEEK_OF_YEAR = 'week-of-year',
	WEEK_OF_MONTH = 'week-of-month',
	DAY_OF_MONTH = 'day-of-month',
	DAY_OF_WEEK = 'day-of-week',
	CASE_THEN = 'case-then'


class ComputedParameter(Parameter):
	kind: ParameterKind.COMPUTED = ParameterKind.COMPUTED
	type: ParameterComputeType = ParameterComputeType.NONE
	parameters: List[Parameter] = []

	def __setattr__(self, name, value):
		if name == 'parameters':
			super().__setattr__(name, construct_parameters(value))
		else:
			super().__setattr__(name, value)


class ParameterCondition(DataModel, BaseModel):
	pass


class ParameterJointType(str, Enum):
	AND = 'and',
	OR = 'or'


class ParameterJoint(ParameterCondition, BaseModel):
	jointType: ParameterJointType = ParameterJointType.AND
	filters: List[ParameterCondition] = []

	def __setattr__(self, name, value):
		if name == 'filters':
			super().__setattr__(name, construct_parameter_conditions(value))
		else:
			super().__setattr__(name, value)


class ParameterExpressionOperator(str, Enum):
	EMPTY = 'empty',
	NOT_EMPTY = 'not-empty',
	EQUALS = 'equals',
	NOT_EQUALS = 'not-equals',
	LESS = 'less',
	LESS_EQUALS = 'less-equals',
	MORE = 'more',
	MORE_EQUALS = 'more-equals',
	IN = 'in',
	NOT_IN = 'not-in',


class ParameterExpression(ParameterCondition, BaseModel):
	left: Parameter = None
	operator: ParameterExpressionOperator = ParameterExpressionOperator.EQUALS
	right: Parameter = None

	def __setattr__(self, name, value):
		if name == 'left' or name == 'right':
			super().__setattr__(name, construct_parameter(value))
		else:
			super().__setattr__(name, value)


def construct_parameter(parameter: Optional[Union[dict, Parameter]]) -> Optional[Parameter]:
	if parameter is None:
		return None
	elif isinstance(parameter, Parameter):
		return parameter
	else:
		kind = parameter.get('kind')
		if kind == ParameterKind.TOPIC:
			return TopicFactorParameter(**parameter)
		elif kind == ParameterKind.CONSTANT:
			return ConstantParameter(**parameter)
		elif kind == ParameterKind.COMPUTED:
			return ComputedParameter(**parameter)
		else:
			raise Exception(f'Parameter kind[{kind}] cannot be recognized.')


def construct_parameters(parameters: Optional[list]) -> Optional[List[Parameter]]:
	if parameters is None:
		return None
	return ArrayHelper(parameters).map(lambda x: construct_parameter(x)).to_list()


def construct_parameter_condition(condition: Optional[Union[dict, ParameterCondition]]) -> Optional[ParameterCondition]:
	if condition is None:
		return None
	elif isinstance(condition, ParameterJoint):
		return condition
	elif isinstance(condition, ParameterExpression):
		return condition
	else:
		joint_type = condition.get('jointType')
		if joint_type is None:
			return ParameterExpression(**condition)
		else:
			return ParameterJoint(**condition)


def construct_parameter_conditions(conditions: Optional[list]) -> Optional[List[ParameterCondition]]:
	if conditions is None:
		return None
	return ArrayHelper(conditions).map(lambda x: construct_parameter_condition(x)).to_list()


def construct_parameter_joint(joint: Optional[Union[dict, ParameterJoint]]) -> Optional[ParameterJoint]:
	if joint is None:
		return None
	elif isinstance(joint, ParameterJoint):
		return joint
	else:
		return ParameterJoint(**joint)



from .pipeline_monitor_log import MonitorAlarmAction, MonitorCopyToMemoryAction, MonitorDeleteAction, \
	MonitorLogAction, MonitorLogActionId, MonitorLogStage, MonitorLogStatus, MonitorLogUnit, MonitorReadAction, \
	MonitorWriteAction, MonitorWriteToExternalAction, PipelineMonitorLog, PipelineMonitorLogCriteria, \
	PipelineMonitorLogId
from .pipeline_monitor_pipelines import ask_pipeline_monitor_pipelines
from .pipeline_monitor_topics import ask_pipeline_monitor_topics
from .pipeline_trigger_data import PipelineTriggerData, PipelineTriggerDataWithPAT, PipelineTriggerResult, \
	PipelineTriggerTraceId, TopicDataColumnNames



from enum import Enum
from typing import Dict, List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import DataSourceId, OptimisticLock, TenantBasedTuple, TopicId
from watchmen_utilities import ArrayHelper
from .factor import Factor


class TopicKind(str, Enum):
	SYSTEM = 'system',
	BUSINESS = 'business'
	SYNONYM = 'synonym'


class TopicType(str, Enum):
	RAW = 'raw',
	META = 'meta',
	DISTINCT = 'distinct',
	AGGREGATE = 'aggregate',
	TIME = 'time',
	RATIO = 'ratio'


def construct_factor(factor: Union[Factor, Dict]) -> Optional[Factor]:
	if factor is None:
		return None
	elif isinstance(factor, Factor):
		return factor
	else:
		return Factor(**factor)


def construct_factors(factors: Optional[List[Union[Factor, Dict]]]) -> Optional[List[Factor]]:
	if factors is None:
		return None
	else:
		return ArrayHelper(factors).map(lambda x: construct_factor(x)).to_list()


class Topic(TenantBasedTuple, OptimisticLock, BaseModel):
	topicId: TopicId = None
	name: str = None
	type: TopicType = TopicType.DISTINCT
	kind: TopicKind = TopicKind.BUSINESS
	dataSourceId: DataSourceId = None
	factors: List[Factor] = []
	description: str = None

	def __setattr__(self, name, value):
		if name == 'factors':
			super().__setattr__(name, construct_factors(value))
		else:
			super().__setattr__(name, value)


def is_raw_topic(topic: Topic) -> bool:
	return topic.type == TopicType.RAW


def is_aggregation_topic(topic: Topic) -> bool:
	topic_type = topic.type
	return topic_type == TopicType.AGGREGATE or topic_type == TopicType.RATIO or topic_type == TopicType.TIME



from enum import Enum

from pydantic import BaseModel

from watchmen_model.common import EnumId, FactorId, Storable


class FactorType(str, Enum):
	SEQUENCE = 'sequence',

	NUMBER = 'number',
	UNSIGNED = 'unsigned',  # 0 & positive

	TEXT = 'text',

	# address
	ADDRESS = 'address',
	CONTINENT = 'continent',
	REGION = 'region',
	COUNTRY = 'country',
	PROVINCE = 'province',
	CITY = 'city',
	DISTRICT = 'district',
	ROAD = 'road',
	COMMUNITY = 'community',
	FLOOR = 'floor',
	RESIDENCE_TYPE = 'residence-type',
	RESIDENTIAL_AREA = 'residential-area',

	# contact electronic
	EMAIL = 'email',
	PHONE = 'phone',
	MOBILE = 'mobile',
	FAX = 'fax',

	# date time related
	DATETIME = 'datetime',  # YYYY-MM-DD HH:mm:ss
	FULL_DATETIME = 'full-datetime',  # YYYY-MM-DD HH:mm:ss.SSS
	DATE = 'date',  # YYYY-MM-DD
	TIME = 'time',  # HH:mm:ss
	YEAR = 'year',  # 4 digits
	HALF_YEAR = 'half-year',  # 1: first half, 2: second half
	QUARTER = 'quarter',  # 1 - 4
	MONTH = 'month',  # 1 - 12
	HALF_MONTH = 'half-month',  # 1: first half, 2: second half
	TEN_DAYS = 'ten-days',  # 1, 2, 3
	WEEK_OF_YEAR = 'week-of-year',  # 0 (the partial week that precedes the first Sunday of the year) - 53 (leap year)
	WEEK_OF_MONTH = 'week-of-month',  # 0 (the partial week that precedes the first Sunday of the year) - 5
	HALF_WEEK = 'half-week',  # 1: first half, 2: second half
	DAY_OF_MONTH = 'day-of-month',  # 1 - 31, according to month/year
	DAY_OF_WEEK = 'day-of-week',  # 1 (Sunday) - 7 (Saturday)
	DAY_KIND = 'day-kind',  # 1: workday, 2: weekend, 3: holiday
	HOUR = 'hour',  # 0 - 23
	HOUR_KIND = 'hour-kind',  # 1: work time, 2: off hours, 3: sleeping time
	MINUTE = 'minute',  # 0 - 59
	SECOND = 'second',  # 0 - 59
	MILLISECOND = 'millisecond',  # 0 - 999
	AM_PM = 'am-pm',  # 1, 2

	# individual
	GENDER = 'gender',
	OCCUPATION = 'occupation',
	DATE_OF_BIRTH = 'date-of-birth',  # YYYY-MM-DD
	AGE = 'age',
	ID_NO = 'id-no',
	RELIGION = 'religion',
	NATIONALITY = 'nationality',

	# organization
	BIZ_TRADE = 'biz-trade',
	BIZ_SCALE = 'biz-scale',

	BOOLEAN = 'boolean',

	ENUM = 'enum',

	OBJECT = 'object',
	ARRAY = 'array'


class FactorIndexGroup(str, Enum):
	INDEX_1 = 'i-1',
	INDEX_2 = 'i-2',
	INDEX_3 = 'i-3',
	INDEX_4 = 'i-4',
	INDEX_5 = 'i-5',
	INDEX_6 = 'i-6',
	INDEX_7 = 'i-7',
	INDEX_8 = 'i-8',
	INDEX_9 = 'i-9',
	INDEX_10 = 'i-10',
	UNIQUE_INDEX_1 = 'u-1',
	UNIQUE_INDEX_2 = 'u-2',
	UNIQUE_INDEX_3 = 'u-3',
	UNIQUE_INDEX_4 = 'u-4',
	UNIQUE_INDEX_5 = 'u-5',
	UNIQUE_INDEX_6 = 'u-6',
	UNIQUE_INDEX_7 = 'u-7',
	UNIQUE_INDEX_8 = 'u-8',
	UNIQUE_INDEX_9 = 'u-9',
	UNIQUE_INDEX_10 = 'u-10',


class FactorEncryptMethod(str, Enum):
	NONE = 'none',
	AES256_PKCS5_PADDING = 'AES256-PKCS5-PADDING',
	MD5 = 'MD5',
	SHA256 = 'SHA256',
	MASK_MAIL = 'MASK-MAIL',
	MASK_CENTER_3 = 'MASK-CENTER-3',
	MASK_CENTER_5 = 'MASK-CENTER-5',
	MASK_LAST_3 = 'MASK-LAST-3',
	MASK_LAST_6 = 'MASK-LAST-6',
	MASK_DAY = 'MASK-DAY',
	MASK_MONTH = 'MASK-MONTH',
	MASK_MONTH_DAY = 'MASK-MONTH-DAY'


class Factor(Storable, BaseModel):
	factorId: FactorId = None
	type: FactorType = None
	name: str = None
	enumId: EnumId = None
	label: str = None
	description: str = None
	defaultValue: str = None
	flatten: bool = False
	indexGroup: FactorIndexGroup = None
	encrypt: FactorEncryptMethod = None
	precision: str = None



from typing import TypeVar

TenantId = TypeVar('TenantId', bound=str)

DataSourceId = TypeVar('DataSourceId', bound=str)
ExternalWriterId = TypeVar('ExternalWriterId', bound=str)

UserId = TypeVar('UserId', bound=str)
UserGroupId = TypeVar('UserGroupId', bound=str)

EnumId = TypeVar('EnumId', bound=str)
EnumItemId = TypeVar('EnumItemId', bound=str)
FactorId = TypeVar('FactorId', bound=str)
TopicId = TypeVar('TopicId', bound=str)
PipelineActionId = TypeVar('PipelineActionId', bound=str)
PipelineUnitId = TypeVar('PipelineUnitId', bound=str)
PipelineStageId = TypeVar('PipelineStageId', bound=str)
PipelineId = TypeVar('PipelineId', bound=str)
PipelineGraphicId = TypeVar('PipelineGraphicId', bound=str)

ReportFunnelId = TypeVar('ReportFunnelId', bound=str)
ReportId = TypeVar('ReportId', bound=str)
SubjectDatasetColumnId = TypeVar('SubjectDatasetColumnId', bound=str)
SubjectId = TypeVar('SubjectId', bound=str)
SpaceId = TypeVar('SpaceId', bound=str)
ConnectedSpaceId = TypeVar('ConnectedSpaceId', bound=str)
DashboardId = TypeVar('DashboardId', bound=str)

BucketId = TypeVar('BucketId', bound=str)
IndicatorId = TypeVar('IndicatorId', bound=str)
InspectionId = TypeVar('InspectionId', bound=str)
AchievementId = TypeVar('AchievementId', bound=str)
ObjectiveAnalysisId = TypeVar('ObjectiveAnalysisId', bound=str)

PatId = TypeVar('PatId', bound=str)



from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, List, Optional, TypeVar, Union

from pydantic import BaseModel

from watchmen_model.admin import DeleteTopicActionType, PipelineActionType, ReadTopicActionType, SystemActionType, \
	WriteTopicActionType
from watchmen_model.common import DataModel, Pageable, PipelineActionId, PipelineId, PipelineStageId, PipelineUnitId, \
	TenantId, TopicId
from watchmen_utilities import ArrayHelper
from .pipeline_trigger_data import PipelineTriggerTraceId


class MonitorLogStatus(str, Enum):
	DONE = 'DONE',  # even step is ignored by prerequisite is false, it is treated as DONE
	IGNORED = 'IGNORED',  # step never be touched
	ERROR = 'ERROR',  # exception occurred


class StandardMonitorLog(DataModel, BaseModel):
	status: MonitorLogStatus
	startTime: Optional[datetime] = None  # keep none when step is ignored
	spentInMills: Optional[int] = 0  # keep 0 when step is ignored
	error: Optional[str]  # if status is ERROR


class ConditionalMonitorLog(StandardMonitorLog):
	prerequisite: bool  # result of prerequisite, True when it is not defined
	prerequisiteDefinedAs: Optional[Any] = None  # definition of prerequisite


MonitorLogActionId = TypeVar('MonitorLogActionId', bound=str)


class MonitorLogAction(StandardMonitorLog):
	uid: MonitorLogActionId
	actionId: PipelineActionId
	type: PipelineActionType
	insertCount: int = 0
	updateCount: int = 0
	deleteCount: int = 0
	definedAs: Optional[Any] = None  # definition of action
	"""
	touched value, 
	for deletion, update and insert, always be list of dict
	for read-exists, bool,
	for read-factor, no arithmetic, Any, depends on factor type
	for read-factor, arithmetic, Decimal
	for read-row, dict
	for read-rows, list of dict
	"""
	touched: Optional[Union[List[Union[Dict[str, Any], Any]], bool, Any, Decimal, Dict[str, Any]]] = None


class MonitorLogFindByAction(MonitorLogAction):
	findBy: Optional[Any] = None  # runtime describing of find by


class MonitorReadAction(MonitorLogFindByAction):
	type: ReadTopicActionType


class MonitorWriteAction(MonitorLogFindByAction):
	type: WriteTopicActionType


class MonitorDeleteAction(MonitorLogFindByAction):
	type: DeleteTopicActionType


class MonitorAlarmAction(MonitorLogAction, ConditionalMonitorLog):
	type: SystemActionType = SystemActionType.ALARM


class MonitorCopyToMemoryAction(MonitorLogAction):
	type: SystemActionType = SystemActionType.COPY_TO_MEMORY


class MonitorWriteToExternalAction(MonitorLogAction):
	type: SystemActionType = SystemActionType.WRITE_TO_EXTERNAL


def is_read_action(action_type: PipelineActionType) -> bool:
	return \
		action_type == ReadTopicActionType.READ_ROW \
		or action_type == ReadTopicActionType.READ_FACTOR \
		or action_type == ReadTopicActionType.EXISTS \
		or action_type == ReadTopicActionType.READ_ROWS \
		or action_type == ReadTopicActionType.READ_FACTORS


def is_write_action(action_type: PipelineActionType) -> bool:
	return \
		action_type == WriteTopicActionType.INSERT_ROW \
		or action_type == WriteTopicActionType.MERGE_ROW \
		or action_type == WriteTopicActionType.INSERT_OR_MERGE_ROW \
		or action_type == WriteTopicActionType.WRITE_FACTOR


def is_delete_action(action_type: PipelineActionType) -> bool:
	return \
		action_type == DeleteTopicActionType.DELETE_ROW \
		or action_type == DeleteTopicActionType.DELETE_ROWS


def construct_action(action: Optional[Union[dict, MonitorLogAction]]) -> Optional[MonitorLogAction]:
	if action is None:
		return None
	elif isinstance(action, MonitorLogAction):
		return action
	else:
		action_type = action.get('type')
		if action_type == SystemActionType.ALARM:
			return MonitorAlarmAction(**action)
		elif action_type == SystemActionType.COPY_TO_MEMORY:
			return MonitorCopyToMemoryAction(**action)
		elif action_type == SystemActionType.WRITE_TO_EXTERNAL:
			return MonitorWriteToExternalAction(**action)
		elif is_read_action(action_type):
			return MonitorReadAction(**action)
		elif is_write_action(action_type):
			return MonitorWriteAction(**action)
		elif is_delete_action(action_type):
			return MonitorDeleteAction(**action)
		else:
			raise Exception(f'Pipeline action type[{action_type}] cannot be recognized.')


def construct_actions(actions: Optional[list] = None) -> Optional[List[MonitorLogAction]]:
	if actions is None:
		return None
	else:
		return ArrayHelper(actions).map(lambda x: construct_action(x)).to_list()


class MonitorLogUnit(ConditionalMonitorLog):
	unitId: PipelineUnitId
	name: str
	loopVariableName: Optional[str] = None
	loopVariableValue: Optional[Any] = None
	actions: List[MonitorLogAction]

	def __setattr__(self, name, value):
		if name == 'actions':
			super().__setattr__(name, construct_actions(value))
		else:
			super().__setattr__(name, value)


def construct_unit(unit: Optional[Union[dict, MonitorLogUnit]]) -> Optional[MonitorLogUnit]:
	if unit is None:
		return None
	elif isinstance(unit, MonitorLogUnit):
		return unit
	else:
		return MonitorLogUnit(**unit)


def construct_units(units: Optional[list] = None) -> Optional[List[MonitorLogUnit]]:
	if units is None:
		return None
	else:
		return ArrayHelper(units).map(lambda x: construct_unit(x)).to_list()


class MonitorLogStage(ConditionalMonitorLog):
	stageId: PipelineStageId
	name: str
	units: List[MonitorLogUnit]

	def __setattr__(self, name, value):
		if name == 'units':
			super().__setattr__(name, construct_units(value))
		else:
			super().__setattr__(name, value)


def construct_stage(stage: Optional[Union[dict, MonitorLogStage]]) -> Optional[MonitorLogStage]:
	if stage is None:
		return None
	elif isinstance(stage, MonitorLogStage):
		return stage
	else:
		return MonitorLogStage(**stage)


def construct_stages(stages: Optional[list] = None) -> Optional[List[MonitorLogStage]]:
	if stages is None:
		return None
	else:
		return ArrayHelper(stages).map(lambda x: construct_stage(x)).to_list()


PipelineMonitorLogId = TypeVar('PipelineMonitorLogId', bound=str)


class PipelineMonitorLog(ConditionalMonitorLog):
	uid: PipelineMonitorLogId
	traceId: PipelineTriggerTraceId
	pipelineId: PipelineId
	topicId: TopicId
	dataId: int
	oldValue: Any
	newValue: Any
	stages: List[MonitorLogStage]

	def __setattr__(self, name, value):
		if name == 'stages':
			super().__setattr__(name, construct_stages(value))
		else:
			super().__setattr__(name, value)


class PipelineMonitorLogCriteria(Pageable):
	topicId: TopicId = None
	pipelineId: PipelineId = None
	startDate: str = None
	endDate: str = None
	status: MonitorLogStatus = None
	traceId: PipelineTriggerTraceId = None
	tenantId: TenantId = None



from datetime import date, datetime, time
from typing import List, Union

from pydantic import BaseModel

DataResultSetCell = Union[str, int, float, bool, datetime, date, time, None]
DataResultSetRow = List[DataResultSetCell]
DataResultSet = List[DataResultSetRow]


class DataResult(BaseModel):
	columns: List[str]
	data: DataResultSet



from typing import List

from watchmen_model.admin import Factor, FactorIndexGroup, FactorType, Topic, TopicKind, TopicType


def ask_pipeline_monitor_topics() -> List[Topic]:
	# TODO define all pipeline monitor topics
	return [
		Topic(
			name='raw_pipeline_monitor_log',
			kind=TopicKind.SYSTEM,
			type=TopicType.RAW,
			factors=[
				Factor(factorId='rpml-f-1', name='uid', type=FactorType.TEXT),
				Factor(
					factorId='rpml-f-2', name='traceId', type=FactorType.TEXT,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_1, precision='50'),
				Factor(
					factorId='rpml-f-3', name='pipelineId', type=FactorType.TEXT,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_2, precision='50'),
				Factor(
					factorId='rpml-f-4', name='topicId', type=FactorType.TEXT,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_3, precision='50'),
				Factor(factorId='rpml-f-5', name='prerequisite', type=FactorType.BOOLEAN),
				Factor(factorId='rpml-f-6', name='prerequisiteDefinedAs', type=FactorType.OBJECT),
				Factor(factorId='rpml-f-7', name='status', type=FactorType.TEXT, flatten=True),
				Factor(factorId='rpml-f-8', name='startTime', type=FactorType.FULL_DATETIME, flatten=True),
				Factor(factorId='rpml-f-9', name='spentInMills', type=FactorType.UNSIGNED, flatten=True),
				Factor(factorId='rpml-f-10', name='error', type=FactorType.TEXT),
				Factor(factorId='rpml-f-11', name='oldValue', type=FactorType.OBJECT),
				Factor(factorId='rpml-f-12', name='newValue', type=FactorType.OBJECT),
				Factor(factorId='rpml-f-13', name='stages', type=FactorType.ARRAY),
				Factor(factorId='rpml-f-14', name='stages.stageId', type=FactorType.TEXT),
				Factor(factorId='rpml-f-15', name='stages.name', type=FactorType.TEXT),
				Factor(factorId='rpml-f-16', name='stages.prerequisite', type=FactorType.BOOLEAN),
				Factor(factorId='rpml-f-17', name='stages.prerequisiteDefinedAs', type=FactorType.OBJECT),
				Factor(factorId='rpml-f-18', name='stages.status', type=FactorType.TEXT),
				Factor(factorId='rpml-f-19', name='stages.startTime', type=FactorType.FULL_DATETIME),
				Factor(factorId='rpml-f-20', name='stages.spentInMills', type=FactorType.UNSIGNED),
				Factor(factorId='rpml-f-21', name='stages.error', type=FactorType.TEXT),
				Factor(factorId='rpml-f-22', name='stages.units', type=FactorType.ARRAY),
				Factor(factorId='rpml-f-23', name='stages.units.unitId', type=FactorType.TEXT),
				Factor(factorId='rpml-f-24', name='stages.units.name', type=FactorType.TEXT),
				Factor(factorId='rpml-f-25', name='stages.units.prerequisite', type=FactorType.BOOLEAN),
				Factor(factorId='rpml-f-26', name='stages.units.prerequisiteDefinedAs', type=FactorType.OBJECT),
				Factor(factorId='rpml-f-27', name='stages.units.loopVariableName', type=FactorType.TEXT),
				Factor(factorId='rpml-f-28', name='stages.units.loopVariableValue', type=FactorType.ARRAY),
				Factor(factorId='rpml-f-29', name='stages.units.status', type=FactorType.TEXT),
				Factor(factorId='rpml-f-30', name='stages.units.startTime', type=FactorType.FULL_DATETIME),
				Factor(factorId='rpml-f-31', name='stages.units.spentInMills', type=FactorType.UNSIGNED),
				Factor(factorId='rpml-f-32', name='stages.units.error', type=FactorType.TEXT),
				Factor(factorId='rpml-f-33', name='stages.units.actions', type=FactorType.ARRAY),
				Factor(factorId='rpml-f-34', name='stages.units.actions.uid', type=FactorType.TEXT),
				Factor(factorId='rpml-f-35', name='stages.units.actions.actionId', type=FactorType.TEXT),
				Factor(factorId='rpml-f-36', name='stages.units.actions.type', type=FactorType.TEXT),
				Factor(factorId='rpml-f-37', name='stages.units.actions.insertCount', type=FactorType.UNSIGNED),
				Factor(factorId='rpml-f-38', name='stages.units.actions.updateCount', type=FactorType.UNSIGNED),
				Factor(factorId='rpml-f-39', name='stages.units.actions.deleteCount', type=FactorType.UNSIGNED),
				Factor(factorId='rpml-f-40', name='stages.units.actions.definedAs', type=FactorType.OBJECT),
				Factor(factorId='rpml-f-41', name='stages.units.actions.findBy', type=FactorType.OBJECT),
				Factor(factorId='rpml-f-42', name='stages.units.actions.touched', type=FactorType.OBJECT),
				Factor(factorId='rpml-f-43', name='stages.units.actions.status', type=FactorType.TEXT),
				Factor(factorId='rpml-f-44', name='stages.units.actions.startTime', type=FactorType.FULL_DATETIME),
				Factor(factorId='rpml-f-45', name='stages.units.actions.spentInMills', type=FactorType.UNSIGNED),
				Factor(factorId='rpml-f-46', name='stages.units.actions.error', type=FactorType.TEXT),
				Factor(
					factorId='rpml-f-47', name='dataId', type=FactorType.NUMBER,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_4, precision='20'),
			],
			description='Pipeline monitor log raw topic.'
		)
	]



from pydantic import BaseModel

from .model import DataModel


class GraphicPosition(DataModel, BaseModel):
	x: float = 0
	y: float = 0


class GraphicSize(DataModel, BaseModel):
	width: float = 0
	height: float = 0


class GraphicRect(GraphicPosition, GraphicSize):
	pass



from pydantic import BaseModel

from watchmen_model.common import DashboardId, LastVisit, UserBasedTuple


class LastSnapshot(UserBasedTuple, LastVisit, BaseModel):
	language: str = None
	lastDashboardId: DashboardId = None
	adminDashboardId: DashboardId = None
	favoritePin: bool = False



from watchmen_model.common import construct_parameter_joint
from .pipeline_action import AggregateArithmeticHolder, FindBy, FromFactor, FromTopic, MemoryWriter, ReadTopicActionType


class ReadTopicAction(FromTopic, MemoryWriter, FindBy):
	type: ReadTopicActionType

	def __setattr__(self, name, value):
		if name == 'by':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


class ReadRowAction(ReadTopicAction):
	type: ReadTopicActionType = ReadTopicActionType.READ_ROW


class ReadRowsAction(ReadTopicAction):
	type: ReadTopicActionType = ReadTopicActionType.READ_ROWS


class ReadFactorAction(FromFactor, ReadTopicAction, AggregateArithmeticHolder):
	type: ReadTopicActionType = ReadTopicActionType.READ_FACTOR


class ReadFactorsAction(FromFactor, ReadTopicAction):
	type: ReadTopicActionType = ReadTopicActionType.READ_FACTORS


class ExistsAction(ReadTopicAction):
	type: ReadTopicActionType = ReadTopicActionType.EXISTS



from enum import Enum

from watchmen_model.common import construct_parameter, construct_parameter_joint, ExternalWriterId, Parameter
from .conditional import Conditional
from .pipeline_action import MemoryWriter, PipelineAction, SystemActionType


class AlarmActionSeverity(str, Enum):
	LOW = 'low',
	MEDIUM = 'medium',
	HIGH = 'high',
	CRITICAL = 'critical',


class AlarmAction(PipelineAction, Conditional):
	type: SystemActionType.ALARM = SystemActionType.ALARM
	severity: AlarmActionSeverity = AlarmActionSeverity.MEDIUM
	message: str = None

	def __setattr__(self, name, value):
		if name == 'on':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


class CopyToMemoryAction(MemoryWriter):
	"""
	copy something to memory variable
	"""
	type: SystemActionType = SystemActionType.COPY_TO_MEMORY
	source: Parameter = None

	def __setattr__(self, name, value):
		if name == 'source':
			super().__setattr__(name, construct_parameter(value))
		else:
			super().__setattr__(name, value)


class WriteToExternalAction(PipelineAction):
	type: SystemActionType = SystemActionType.WRITE_TO_EXTERNAL
	externalWriterId: ExternalWriterId = None
	eventCode: str = None



from datetime import datetime

from .model import DataModel
from .tuple_ids import UserId


class Storable(DataModel):
	pass


class Auditable(Storable):
	createdAt: datetime = None
	createdBy: UserId = None
	lastModifiedAt: datetime = None
	lastModifiedBy: UserId = None


class OptimisticLock(Storable):
	version: int = 1


class LastVisit(Storable):
	lastVisitTime: datetime = datetime.now().replace(tzinfo=None, microsecond=0)



from .achievement import Achievement, AchievementIndicator, AchievementTimeRangeType, \
	MANUAL_COMPUTE_ACHIEVEMENT_INDICATOR_ID, ManualComputeAchievementIndicator
from .bucket import Bucket, BucketSegment, BucketType, CategoryMeasureBucket, CategorySegment, CategorySegmentsHolder, \
	CategorySegmentValue, EnumMeasureBucket, MeasureBucket, NumericSegmentsHolder, NumericSegmentValue, \
	NumericValueBucket, NumericValueMeasureBucket, NumericValueSegment, OtherCategorySegmentValue, \
	RangeBucketValueIncluding
from .indicator import Indicator, IndicatorAggregateArithmetic, IndicatorBaseOn, IndicatorFilter, RelevantIndicator, \
	RelevantIndicatorType
from .indicator_criteria import IndicatorCriteria, IndicatorCriteriaOnBucket, IndicatorCriteriaOnExpression, \
	IndicatorCriteriaOperator
from .inspection import Inspection, InspectionAmPmRange, InspectionDayKindRange, InspectionDayOfMonthRange, \
	InspectionDayOfWeekRange, InspectionHalfMonthRange, InspectionHalfWeekRange, InspectionHalfYearRange, \
	InspectionHourKindRange, InspectionHourRange, InspectionMonthRange, InspectionQuarterRange, \
	InspectionTenDaysRange, InspectionTimeRange, InspectionTimeRangeType, InspectionWeekOfMonthRange, \
	InspectionWeekOfYearRange, InspectionYearRange, InspectMeasureOn
from .measure_method import MeasureMethod
from .objective_analysis import ObjectiveAnalysis, ObjectiveAnalysisPerspective, ObjectiveAnalysisPerspectiveType



from typing import Any

from pydantic import BaseModel

from .chart_basic_structure import ChartTruncation, ChartTruncationHolder
from .chart_basic_style import ChartBorder
from .chart_enums import PredefinedChartColorSeries
from .chart_types import ChartColor


class ChartSettings(ChartTruncationHolder, BaseModel):
	border: ChartBorder = None
	backgroundColor: ChartColor = None

	colorSeries: PredefinedChartColorSeries = PredefinedChartColorSeries.REGULAR

	# noinspection PyMethodMayBeStatic,PyUnusedLocal
	def handle_attr(self, name, value) -> Any:
		return value

	def __setattr__(self, name, value):
		if name == 'truncation':
			if isinstance(value, ChartTruncation):
				super().__setattr__(name, value)
			else:
				super().__setattr__(name, ChartTruncation(**value))
		else:
			super().__setattr__(name, self.handle_attr(name, value))



from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import construct_parameter_joint, DataModel, OptimisticLock, ParameterJoint, SpaceId, \
	TenantBasedTuple, TopicId, UserGroupId
from watchmen_utilities import ArrayHelper


# noinspection PyRedundantParentheses,DuplicatedCode
class AvoidFastApiError:
	joint: ParameterJoint = None


class SpaceFilter(DataModel, AvoidFastApiError, BaseModel):
	topicId: TopicId = None
	enabled: bool = False

	def __setattr__(self, name, value):
		if name == 'joint':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


def construct_filter(a_filter: Optional[Union[dict, SpaceFilter]]) -> Optional[SpaceFilter]:
	if a_filter is None:
		return None
	elif isinstance(a_filter, SpaceFilter):
		return a_filter
	else:
		# noinspection PyArgumentList
		return SpaceFilter(**a_filter)


def construct_filters(filters: Optional[list] = None) -> Optional[List[SpaceFilter]]:
	if filters is None:
		return None
	else:
		return ArrayHelper(filters).map(lambda x: construct_filter(x)).to_list()


class Space(TenantBasedTuple, OptimisticLock, BaseModel):
	spaceId: SpaceId = None
	name: str = None
	description: str = None
	topicIds: List[TopicId] = None
	groupIds: List[UserGroupId] = None
	filters: List[SpaceFilter] = None

	def __setattr__(self, name, value):
		if name == 'filters':
			super().__setattr__(name, construct_filters(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import construct_parameter_joint, OptimisticLock, PipelineId, \
	PipelineStageId, PipelineUnitId, TenantBasedTuple, TopicId
from watchmen_utilities import ArrayHelper
from .conditional import Conditional
from .pipeline_action import DeleteTopicActionType, PipelineAction, ReadTopicActionType, SystemActionType, \
	WriteTopicActionType
from .pipeline_action_delete import DeleteRowAction, DeleteRowsAction
from .pipeline_action_read import ExistsAction, ReadFactorAction, ReadFactorsAction, ReadRowAction, ReadRowsAction
from .pipeline_action_system import AlarmAction, CopyToMemoryAction, WriteToExternalAction
from .pipeline_action_write import InsertOrMergeRowAction, InsertRowAction, MergeRowAction, WriteFactorAction


def construct_action(action: Optional[Union[dict, PipelineAction]]) -> Optional[PipelineAction]:
	if action is None:
		return None
	elif isinstance(action, PipelineAction):
		return action
	else:
		action_type = action.get('type')
		if action_type == SystemActionType.ALARM:
			return AlarmAction(**action)
		elif action_type == SystemActionType.COPY_TO_MEMORY:
			return CopyToMemoryAction(**action)
		elif action_type == SystemActionType.WRITE_TO_EXTERNAL:
			return WriteToExternalAction(**action)
		elif action_type == ReadTopicActionType.READ_ROW:
			return ReadRowAction(**action)
		elif action_type == ReadTopicActionType.READ_FACTOR:
			return ReadFactorAction(**action)
		elif action_type == ReadTopicActionType.EXISTS:
			return ExistsAction(**action)
		elif action_type == ReadTopicActionType.READ_ROWS:
			return ReadRowsAction(**action)
		elif action_type == ReadTopicActionType.READ_FACTORS:
			return ReadFactorsAction(**action)
		elif action_type == WriteTopicActionType.MERGE_ROW:
			return MergeRowAction(**action)
		elif action_type == WriteTopicActionType.INSERT_ROW:
			return InsertRowAction(**action)
		elif action_type == WriteTopicActionType.INSERT_OR_MERGE_ROW:
			return InsertOrMergeRowAction(**action)
		elif action_type == WriteTopicActionType.WRITE_FACTOR:
			return WriteFactorAction(**action)
		elif action_type == DeleteTopicActionType.DELETE_ROW:
			return DeleteRowAction(**action)
		elif action_type == DeleteTopicActionType.DELETE_ROWS:
			return DeleteRowsAction(**action)
		else:
			raise Exception(f'Pipeline action type[{action_type}] cannot be recognized.')


def construct_actions(actions: Optional[list] = None) -> Optional[List[PipelineAction]]:
	if actions is None:
		return None
	else:
		return ArrayHelper(actions).map(lambda x: construct_action(x)).to_list()


class PipelineUnit(Conditional, BaseModel):
	unitId: PipelineUnitId = None
	name: str = None
	loopVariableName: str = None
	do: List[PipelineAction] = []

	def __setattr__(self, name, value):
		if name == 'do':
			super().__setattr__(name, construct_actions(value))
		elif name == 'on':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


def construct_unit(unit: Optional[Union[dict, PipelineUnit]]) -> Optional[PipelineUnit]:
	if unit is None:
		return None
	elif isinstance(unit, PipelineUnit):
		return unit
	else:
		return PipelineUnit(**unit)


def construct_units(units: Optional[list] = None) -> Optional[List[PipelineUnit]]:
	if units is None:
		return None
	else:
		return ArrayHelper(units).map(lambda x: construct_unit(x)).to_list()


class PipelineStage(Conditional, BaseModel):
	stageId: PipelineStageId = None
	name: str = None
	units: List[PipelineUnit] = []

	def __setattr__(self, name, value):
		if name == 'units':
			super().__setattr__(name, construct_units(value))
		elif name == 'on':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


class PipelineTriggerType(str, Enum):
	INSERT = 'insert',
	MERGE = 'merge',
	INSERT_OR_MERGE = 'insert-or-merge',
	DELETE = 'delete',


def construct_stage(stage: Optional[Union[dict, PipelineStage]]) -> Optional[PipelineStage]:
	if stage is None:
		return None
	elif isinstance(stage, PipelineStage):
		return stage
	else:
		return PipelineStage(**stage)


def construct_stages(stages: Optional[list] = None) -> Optional[List[PipelineStage]]:
	if stages is None:
		return None
	else:
		return ArrayHelper(stages).map(lambda x: construct_stage(x)).to_list()


class Pipeline(Conditional, TenantBasedTuple, OptimisticLock, BaseModel):
	pipelineId: PipelineId = None
	topicId: TopicId = None
	name: str = None
	type: PipelineTriggerType = None
	stages: List[PipelineStage] = []
	enabled: bool = None
	validated: bool = None

	def __setattr__(self, name, value):
		if name == 'stages':
			super().__setattr__(name, construct_stages(value))
		elif name == 'on':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)



from .data_source import DataSource, DataSourceParam, DataSourceType
from .external_writer import ExternalWriter, ExternalWriterType
from .key_store import KeyStore
from .tenant import Tenant
from .token import PersonalAccessToken, Token



from typing import List

from watchmen_model.admin import Factor, MappingFactor, Pipeline, PipelineStage, PipelineTriggerType, PipelineUnit, \
	Topic, WriteTopicActionType
from watchmen_model.admin.pipeline_action_write import InsertOrMergeRowAction
from watchmen_model.common import ComputedParameter, ConstantParameter, ParameterComputeType, ParameterExpression, \
	ParameterExpressionOperator, ParameterJoint, ParameterJointType, ParameterKind, TopicFactorParameter
from watchmen_utilities import ArrayHelper


def find_topic(topics: List[Topic], topic_name: str) -> Topic:
	return ArrayHelper(topics).find(lambda x: x.name == topic_name)


def find_factor(topic: Topic, factor_name: str) -> Factor:
	return ArrayHelper(topic.factors).find(lambda x: x.name == factor_name)


def find_source_parameter(topic: Topic, factor_name: str) -> TopicFactorParameter:
	return TopicFactorParameter(
		kind=ParameterKind.TOPIC, topicId=topic.topicId,
		factorId=find_factor(topic, factor_name).factorId
	)


def find_part_of_process_date(topic: Topic, factor_name: str, part: ParameterComputeType) -> ComputedParameter:
	return ComputedParameter(
		kind=ParameterKind.COMPUTED,
		type=part,
		parameters=[find_source_parameter(topic, factor_name)]
	)


def ask_dqc_pipelines(topics: List[Topic]) -> List[Pipeline]:
	# define all dqc pipelines
	topic_raw: Topic = find_topic(topics, 'dqc_raw_rule_result')
	topic_daily: Topic = find_topic(topics, 'dqc_rule_daily')

	return [
		Pipeline(
			name='DQC Pipeline',
			topicId=topic_raw.topicId,
			type=PipelineTriggerType.INSERT_OR_MERGE,
			stages=[PipelineStage(
				stageId='dqcp-s-1',
				name='DQC Pipeline Stage 1',
				units=[PipelineUnit(
					unitId='dqcp-u-1',
					name='DQC Pipeline Unit 1',
					do=[InsertOrMergeRowAction(
						actionId='dqcp-a-1',
						type=WriteTopicActionType.INSERT_OR_MERGE_ROW,
						topicId=topic_daily.topicId,
						mapping=[
							MappingFactor(
								source=find_source_parameter(topic_raw, 'ruleCode'),
								factorId=find_factor(topic_daily, 'ruleCode').factorId),
							MappingFactor(
								source=find_source_parameter(topic_raw, 'topicId'),
								factorId=find_factor(topic_daily, 'topicId').factorId),
							MappingFactor(
								source=find_source_parameter(topic_raw, 'factorId'),
								factorId=find_factor(topic_daily, 'factorId').factorId),
							MappingFactor(
								source=find_part_of_process_date(
									topic_raw, 'processDate', ParameterComputeType.YEAR_OF),
								factorId=find_factor(topic_daily, 'year').factorId),
							MappingFactor(
								source=find_part_of_process_date(
									topic_raw, 'processDate', ParameterComputeType.MONTH_OF),
								factorId=find_factor(topic_daily, 'month').factorId),
							MappingFactor(
								source=find_part_of_process_date(
									topic_raw, 'processDate', ParameterComputeType.DAY_OF_MONTH),
								factorId=find_factor(topic_daily, 'day').factorId),
							MappingFactor(
								source=find_source_parameter(topic_raw, 'processDate'),
								factorId=find_factor(topic_daily, 'processDate').factorId),
							MappingFactor(
								source=ComputedParameter(
									kind=ParameterKind.COMPUTED,
									type=ParameterComputeType.CASE_THEN,
									parameters=[
										ConstantParameter(
											kind=ParameterKind.CONSTANT,
											conditional=True,
											on=ParameterJoint(
												jointType=ParameterJointType.AND,
												filters=[
													ParameterExpression(
														left=find_source_parameter(topic_raw, 'detected'),
														operator=ParameterExpressionOperator.EQUALS,
														right=ConstantParameter(
															kind=ParameterKind.CONSTANT,
															value="true"
														)
													)
												]
											),
											value="1"
										),
										ConstantParameter(
											kind=ParameterKind.CONSTANT,
											value="0"
										)
									]
								),
								factorId=find_factor(topic_daily, 'count').factorId)
						],
						by=ParameterJoint(
							jointType=ParameterJointType.AND,
							filters=[
								ParameterExpression(
									left=find_source_parameter(topic_raw, 'ruleCode'),
									operator=ParameterExpressionOperator.EQUALS,
									right=find_source_parameter(topic_daily, 'ruleCode')
								),
								ParameterExpression(
									left=find_source_parameter(topic_raw, 'topicId'),
									operator=ParameterExpressionOperator.EQUALS,
									right=find_source_parameter(topic_daily, 'topicId')
								),
								ParameterExpression(
									left=find_source_parameter(topic_raw, 'factorId'),
									operator=ParameterExpressionOperator.EQUALS,
									right=find_source_parameter(topic_daily, 'factorId')
								),
								ParameterExpression(
									left=find_part_of_process_date(
										topic_raw, 'processDate', ParameterComputeType.YEAR_OF),
									operator=ParameterExpressionOperator.EQUALS,
									right=find_source_parameter(topic_daily, 'year')
								),
								ParameterExpression(
									left=find_part_of_process_date(
										topic_raw, 'processDate', ParameterComputeType.MONTH_OF),
									operator=ParameterExpressionOperator.EQUALS,
									right=find_source_parameter(topic_daily, 'month')
								),
								ParameterExpression(
									left=find_part_of_process_date(
										topic_raw, 'processDate', ParameterComputeType.DAY_OF_MONTH),
									operator=ParameterExpressionOperator.EQUALS,
									right=find_source_parameter(topic_daily, 'day')
								)
							]
						)
					)]
				)]
			)],
			enabled=True,
			validated=True
		)
	]



from typing import Dict, List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import EnumId, EnumItemId, OptimisticLock, Storable, TenantBasedTuple, TenantId
from watchmen_utilities import ArrayHelper


class EnumItem(Storable, BaseModel):
	itemId: EnumItemId = None
	code: str = None
	label: str = None
	parentCode: str = None
	replaceCode: str = None
	enumId: EnumId = None
	tenantId: TenantId = None


def construct_item(item: Union[EnumItem, Dict]) -> Optional[EnumItem]:
	if item is None:
		return None
	elif isinstance(item, EnumItem):
		return item
	else:
		return EnumItem(**item)


def construct_items(items: Optional[List[Union[EnumItem, Dict]]]) -> Optional[List[EnumItem]]:
	if items is None:
		return None
	else:
		return ArrayHelper(items).map(lambda x: construct_item(x)).to_list()


class Enum(TenantBasedTuple, OptimisticLock, BaseModel):
	enumId: EnumId = None
	name: str = None
	description: str = None
	parentEnumId: EnumId = None
	items: List[EnumItem] = []

	def __setattr__(self, name, value):
		if name == 'items':
			super().__setattr__(name, construct_items(value))
		else:
			super().__setattr__(name, value)



from datetime import datetime
from typing import TypeVar

from pydantic import BaseModel

from watchmen_model.admin import FactorType
from watchmen_model.common import FactorId, Storable, TenantId, TopicId

FactorIndexId = TypeVar('FactorIndexId', bound=str)


# noinspection DuplicatedCode
class FactorIndex(Storable, BaseModel):
	factorIndexId: FactorIndexId = None
	factorId: FactorId = None
	factorType: FactorType = None
	factorName: str = None
	factorLabel: str = None
	factorDescription: str = None
	topicId: TopicId = None
	topicName: str = None
	tenantId: TenantId = None
	createdAt: datetime = None
	lastModifiedAt: datetime = None



from .data_result_set import DataResult, DataResultSet, DataResultSetCell, DataResultSetRow
from .graphic import GraphicPosition, GraphicRect, GraphicSize
from .model import DataModel, SettingsModel
from .pagination import DataPage, Pageable, PageDataCell, PageDataRow, PageDataSet
from .parameter_and_condition import ComputedParameter, ConstantParameter, construct_parameter, \
	construct_parameter_condition, construct_parameter_conditions, construct_parameter_joint, Parameter, \
	ParameterComputeType, ParameterCondition, ParameterExpression, ParameterExpressionOperator, ParameterJoint, \
	ParameterJointType, ParameterKind, TopicFactorParameter, VariablePredefineFunctions
from .storable import Auditable, LastVisit, OptimisticLock, Storable
from .tuple import TenantBasedTuple, Tuple, UserBasedTuple
from .tuple_ids import AchievementId, BucketId, ConnectedSpaceId, DashboardId, DataSourceId, EnumId, EnumItemId, \
	ExternalWriterId, FactorId, IndicatorId, InspectionId, ObjectiveAnalysisId, PatId, PipelineActionId, \
	PipelineGraphicId, PipelineId, PipelineStageId, PipelineUnitId, ReportFunnelId, ReportId, SpaceId, \
	SubjectDatasetColumnId, SubjectId, TenantId, TopicId, UserGroupId, UserId



from typing import List

from pydantic import BaseModel

from watchmen_model.common import IndicatorId, OptimisticLock, SpaceId, TenantBasedTuple, UserGroupId, UserId


class UserGroup(TenantBasedTuple, OptimisticLock, BaseModel):
	userGroupId: UserGroupId = None
	name: str = None
	description: str = None
	userIds: List[UserId] = None
	spaceIds: List[SpaceId] = None
	indicatorIds: List[IndicatorId] = None



from datetime import datetime
from typing import Any, Dict

from pydantic import BaseModel

from watchmen_model.common import Storable, TenantId, UserId


class KeyStore(Storable, BaseModel):
	tenantId: TenantId = None
	keyType: str = None
	params: Dict[str, Any] = None
	createdAt: datetime = None
	createdBy: UserId = None



from .chart import BarChart, BarChartSettings, BarChartSettingsLabel, BarChartSettingsSeries, BarLabelPosition, \
	BooleanItem, Chart, ColorItem, CountChart, CountChartSettings, CountChartSettingsText, CustomizedChart, \
	CustomizedChartSettings, DefItem, DoughnutChart, DoughnutChartSettings, DropdownItem, DropdownItemOption, \
	EChartsAlignmentHolder, EChartsAxisMinorSplitLineHolder, EChartsAxisSplitLine, EChartsAxisSplitLineHolder, \
	EChartsAxisSplitLineStyle, EChartsBorderHolder, EChartsBorderHolderNoRadius, EChartsBorderOmitRadius, \
	EChartsFontHolder, EChartsGrid, EChartsGridHolder, EChartsGridPositionOnly, EChartsHorizontalAlignment, \
	EChartsLegend, EChartsLegendHolder, EChartsLegendOrient, EChartsPosition, EChartsPositionHolder, \
	EchartsScriptHolder, EchartsScriptsVars, EChartsSettings, EChartsTitle, EChartsTitleHolder, EChartsTitleText, \
	EChartsToolbox, EChartsToolboxHolder, EChartsToolboxOrient, EChartsVerticalAlignment, EChartsXAxis, \
	EChartsXAxisHolder, EChartsXAxisLabel, EChartsXAxisName, EChartsXAxisNameLocation, EChartsXAxisPosition, \
	EChartsXAxisType, EChartsYAxis, EChartsYAxisHolder, EChartsYAxisLabel, EChartsYAxisName, EChartsYAxisNameLocation, \
	EChartsYAxisPosition, EChartsYAxisType, InputItem, ItemType, LineChart, LineChartSettings, \
	LineChartSettingsSeries, MapChart, MapChartRegion, MapChartSettings, MapChartSettingsSeries, NightingaleChart, \
	NightingaleChartSettings, NumberItem, PercentageItem, PieChart, PieChartSettings, PieChartSettingsLabel, \
	PieChartSettingsSeries, PieLabelAlignTo, PieLabelPosition, PieRoseType, ScatterChart, ScatterChartSettings, \
	SectionItem, SunburstChart, SunburstChartSettings, SunburstChartSettingsSeries, TextItem, TreeChart, \
	TreeChartSettings, TreeChartSettingsSeries, TreeLayout, TreemapChart, TreemapChartSettings, \
	TreemapChartSettingsSeries, TreeOrient
from .chart_basic_structure import ChartTruncation, ChartTruncationHolder, ChartTruncationType
from .chart_basic_style import ChartBorder, ChartFont
from .chart_enums import ChartBorderStyle, ChartFontStyle, ChartFontWeight, PredefinedChartColorSeries
from .chart_settings import ChartSettings
from .chart_types import ChartColor, ChartType



from pydantic import BaseModel

from watchmen_model.common import Auditable, ConnectedSpaceId, LastVisit, SpaceId, UserBasedTuple


class ConnectedSpace(UserBasedTuple, Auditable, LastVisit, BaseModel):
	connectId: ConnectedSpaceId = None
	spaceId: SpaceId = None
	name: str = None
	isTemplate: bool = False



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import AchievementId, Auditable, DataModel, IndicatorId, UserBasedTuple
from watchmen_utilities import ArrayHelper
from .indicator import IndicatorAggregateArithmetic
from .indicator_criteria import construct_indicator_criteria_list, IndicatorCriteria


class AchievementIndicator(DataModel, BaseModel):
	indicatorId: IndicatorId = None
	name: str = None
	aggregateArithmetic: IndicatorAggregateArithmetic = None
	formula: str = None
	includeInFinalScore: bool = True
	criteria: List[IndicatorCriteria] = []
	variableName: str = None

	def __setattr__(self, name, value):
		if name == 'criteria':
			super().__setattr__(name, construct_indicator_criteria_list(value))
		else:
			super().__setattr__(name, value)


MANUAL_COMPUTE_ACHIEVEMENT_INDICATOR_ID = '-1'


class ManualComputeAchievementIndicator(AchievementIndicator):
	"""
	for manual compute indicator,
	1. indicatorId fixed as {@link MANUAL_COMPUTE_ACHIEVEMENT_INDICATOR_ID},
	2. aggregateArithmetics fixed as {@link IndicatorAggregateArithmetic#MAX}, will be ignored anyway in runtime
	3. criteria fixed as zero length array, will be ignored anyway in runtime
	"""
	indicatorId: IndicatorId = MANUAL_COMPUTE_ACHIEVEMENT_INDICATOR_ID
	aggregateArithmetic: IndicatorAggregateArithmetic = IndicatorAggregateArithmetic.MAX


class AchievementTimeRangeType(str, Enum):
	YEAR = 'year',
	MONTH = 'month'


def construct_indicator(indicator: Optional[Union[dict, AchievementIndicator]]) -> Optional[AchievementIndicator]:
	if indicator is None:
		return None
	elif isinstance(indicator, AchievementIndicator):
		return indicator
	else:
		indicator_id = indicator.get('indicatorId')
		if indicator_id == MANUAL_COMPUTE_ACHIEVEMENT_INDICATOR_ID:
			return ManualComputeAchievementIndicator(**indicator)
		else:
			return AchievementIndicator(**indicator)


def construct_indicators(indicators: Optional[list] = None) -> Optional[List[AchievementIndicator]]:
	if indicators is None:
		return None
	else:
		return ArrayHelper(indicators).map(lambda x: construct_indicator(x)).to_list()


class Achievement(UserBasedTuple, Auditable, BaseModel):
	achievementId: AchievementId = None
	name: str = None
	description: str = None
	timeRangeType: AchievementTimeRangeType = AchievementTimeRangeType.YEAR
	timeRangeYear: str = None
	timeRangeMonth: str = None
	compareWithPreviousTimeRange: bool = False
	finalScoreIsRatio: bool = False
	indicators: List[AchievementIndicator] = []

	def __setattr__(self, name, value):
		if name == 'indicators':
			super().__setattr__(name, construct_indicators(value))
		else:
			super().__setattr__(name, value)



from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import ConnectedSpaceId, DataModel, GraphicRect, SubjectId, TopicId, UserBasedTuple
from watchmen_utilities import ArrayHelper
from .utils import construct_rect


class TopicGraphic(DataModel, BaseModel):
	topicId: TopicId = None
	rect: GraphicRect = None

	def __setattr__(self, name, value):
		if name == 'rect':
			super().__setattr__(name, construct_rect(value))
		else:
			super().__setattr__(name, value)


class SubjectGraphic(DataModel, BaseModel):
	subjectId: SubjectId = None
	rect: GraphicRect = None

	def __setattr__(self, name, value):
		if name == 'rect':
			super().__setattr__(name, construct_rect(value))
		else:
			super().__setattr__(name, value)


# noinspection DuplicatedCode
def construct_topic(topic: Optional[Union[dict, TopicGraphic]]) -> Optional[TopicGraphic]:
	if topic is None:
		return None
	elif isinstance(topic, TopicGraphic):
		return topic
	else:
		return TopicGraphic(**topic)


def construct_topics(topics: List[Union[dict, TopicGraphic]]) -> List[TopicGraphic]:
	if topics is None:
		return []
	return ArrayHelper(topics).map(lambda x: construct_topic(x)).to_list()


def construct_subject(subject: Optional[Union[dict, SubjectGraphic]]) -> Optional[SubjectGraphic]:
	if subject is None:
		return None
	elif isinstance(subject, SubjectGraphic):
		return subject
	else:
		return SubjectGraphic(**subject)


def construct_subjects(subjects: List[Union[dict, SubjectGraphic]]) -> List[SubjectGraphic]:
	if subjects is None:
		return []
	return ArrayHelper(subjects).map(lambda x: construct_subject(x)).to_list()


class ConnectedSpaceGraphic(UserBasedTuple, BaseModel):
	connectId: ConnectedSpaceId = None
	topics: List[TopicGraphic] = None
	subjects: List[SubjectGraphic] = None

	def __setattr__(self, name, value):
		if name == 'topics':
			super().__setattr__(name, construct_topics(value))
		elif name == 'subjects':
			super().__setattr__(name, construct_subjects(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import AchievementId, Auditable, DataModel, InspectionId, LastVisit, ObjectiveAnalysisId, \
	UserBasedTuple
from watchmen_utilities import ArrayHelper


class ObjectiveAnalysisPerspectiveType(str, Enum):
	INSPECTION = 'inspection',
	ACHIEVEMENT = 'achievement'


class ObjectiveAnalysisPerspective(DataModel, BaseModel):
	perspectiveId: str = None
	description: Optional[str] = None
	type: ObjectiveAnalysisPerspectiveType = ObjectiveAnalysisPerspectiveType.INSPECTION
	relationId: Optional[Union[InspectionId, AchievementId]] = None


def construct_perspective(
		perspective: Optional[Union[dict, ObjectiveAnalysisPerspective]]
) -> Optional[ObjectiveAnalysisPerspective]:
	if perspective is None:
		return None
	elif isinstance(perspective, ObjectiveAnalysisPerspective):
		return perspective
	else:
		return ObjectiveAnalysisPerspective(**perspective)


def construct_perspectives(indicators: Optional[list] = None) -> Optional[List[ObjectiveAnalysisPerspective]]:
	if indicators is None:
		return None
	else:
		return ArrayHelper(indicators).map(lambda x: construct_perspective(x)).to_list()


class ObjectiveAnalysis(UserBasedTuple, Auditable, LastVisit, BaseModel):
	analysisId: ObjectiveAnalysisId = None
	title: str = None
	description: str = None
	perspectives: List[ObjectiveAnalysisPerspective] = []

	def __setattr__(self, name, value):
		if name == 'perspectives':
			super().__setattr__(name, construct_perspectives(value))
		else:
			super().__setattr__(name, value)



from datetime import datetime
from enum import Enum
from typing import Optional, TypeVar

from pydantic import BaseModel

from watchmen_model.common import FactorId, PipelineActionId, PipelineId, PipelineStageId, PipelineUnitId, Storable, \
	TenantId, TopicId

PipelineIndexId = TypeVar('PipelineIndexId', bound=str)


class PipelineIndexRefType(str, Enum):
	DIRECT = 'direct',
	COMPUTED = 'computed'


class PipelineIndex(Storable, BaseModel):
	pipelineIndexId: PipelineIndexId = None
	pipelineId: PipelineId = None
	pipelineName: str = None
	stageId: PipelineStageId = None
	stageName: str = None
	unitId: PipelineUnitId = None
	unitName: str = None
	actionId: PipelineActionId = None
	mappingToTopicId: Optional[TopicId] = None
	mappingToFactorId: Optional[FactorId] = None
	sourceFromTopicId: Optional[TopicId] = None
	sourceFromFactorId: Optional[FactorId] = None
	refType: PipelineIndexRefType = None
	tenantId: TenantId = None
	createdAt: datetime = None
	lastModifiedAt: datetime = None



from typing import Optional, Union

from watchmen_model.common import GraphicRect


def construct_rect(rect: Optional[Union[dict, GraphicRect]]) -> Optional[GraphicRect]:
	if rect is None:
		return None
	elif isinstance(rect, GraphicRect):
		return rect
	else:
		return GraphicRect(**rect)



from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from watchmen_model.common import DataModel, FactorId, TopicId
from .monitor_rule import MonitorRuleCode


class MonitorRuleLog(DataModel, BaseModel):
	ruleCode: MonitorRuleCode = None
	topicId: Optional[TopicId] = None
	factorId: Optional[FactorId] = None
	count: int = None
	lastOccurredTime: datetime = None


class MonitorRuleLogCriteria(DataModel, BaseModel):
	startDate: str = None
	endDate: str = None
	ruleCode: MonitorRuleCode = None
	topicId: TopicId = None
	factorId: FactorId = None



from datetime import date
from typing import List, Optional

from watchmen_model.admin import Factor, FactorIndexGroup, FactorType, Topic, TopicKind, TopicType
from watchmen_model.common import DataModel, FactorId, TopicId
from .monitor_rule import MonitorRuleCode, MonitorRuleSeverity


class MonitorRuleDetected(DataModel):
	ruleCode: MonitorRuleCode
	topicId: TopicId
	topicName: str
	factorId: Optional[FactorId] = None
	factorName: Optional[str]
	detected: bool  # issue detected
	severity: MonitorRuleSeverity
	processDate: date


def ask_dqc_topics() -> List[Topic]:
	# TODO define all dqc topics
	return [
		Topic(
			name='dqc_raw_rule_result',
			kind=TopicKind.SYSTEM,
			type=TopicType.RAW,
			factors=[
				Factor(
					factorId='dra-f-1', name='ruleCode', type=FactorType.TEXT,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_1, precision='50'),
				Factor(
					factorId='dra-f-2', name='topicId', type=FactorType.TEXT,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_2, precision='50'),
				Factor(
					factorId='dra-f-3', name='topicName', type=FactorType.TEXT),
				Factor(
					factorId='dra-f-4', name='factorId', type=FactorType.TEXT,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_3, precision='50'),
				Factor(
					factorId='dra-f-5', name='factorName', type=FactorType.TEXT),
				Factor(factorId='dra-f-6', name='detected', type=FactorType.BOOLEAN, flatten=True),
				Factor(factorId='dra-f-7', name='severity', type=FactorType.TEXT),
				# the start day of date range
				# sunday of weekly; 1st of monthly.
				Factor(
					factorId='dra-f-8', name='processDate', type=FactorType.DATE,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_4)
			],
			description='Topic data monitor by rules, raw topic.'
		),
		Topic(
			name='dqc_rule_daily',
			kind=TopicKind.SYSTEM,
			type=TopicType.DISTINCT,
			factors=[
				Factor(factorId='dra-f-1', name='ruleCode', type=FactorType.TEXT, indexGroup=FactorIndexGroup.INDEX_1),
				Factor(factorId='dra-f-2', name='topicId', type=FactorType.TEXT, indexGroup=FactorIndexGroup.INDEX_2),
				Factor(factorId='dra-f-3', name='factorId', type=FactorType.TEXT, indexGroup=FactorIndexGroup.INDEX_3),
				Factor(factorId='dra-f-4', name='year', type=FactorType.YEAR, indexGroup=FactorIndexGroup.INDEX_4),
				Factor(factorId='dra-f-5', name='month', type=FactorType.MONTH, indexGroup=FactorIndexGroup.INDEX_5),
				Factor(
					factorId='dra-f-6', name='day', type=FactorType.DAY_OF_MONTH, indexGroup=FactorIndexGroup.INDEX_6),
				Factor(
					factorId='dra-f-7', name='processDate', type=FactorType.DATE, indexGroup=FactorIndexGroup.INDEX_7),
				Factor(factorId='dra-f-8', name='count', type=FactorType.UNSIGNED, precision='10')
			],
			description='Topic data monitor by rules, distinct topic.'
		)
	]



from datetime import datetime

from pydantic import BaseModel

from watchmen_model.common import PatId, TenantId, UserBasedTuple


class Token(BaseModel):
	accessToken: str
	tokenType: str
	role: str
	tenantId: TenantId = None


class PersonalAccessToken(UserBasedTuple, BaseModel):
	patId: PatId = None
	token: str = None
	username: str = None
	note: str = None
	expired: datetime = None
	permissions: list = None
	createdAt: datetime = None



from enum import Enum
from typing import Any, List, Literal, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import Auditable, BucketId, DataModel, FactorId, IndicatorId, InspectionId, UserBasedTuple
from watchmen_utilities import ArrayHelper
from .indicator import IndicatorAggregateArithmetic
from .indicator_criteria import construct_indicator_criteria_list, IndicatorCriteria
from .measure_method import MeasureMethod


class InspectMeasureOn(str, Enum):
	NONE = 'none',
	VALUE = 'value',
	OTHER = 'other',


class InspectionTimeRangeType(str, Enum):
	YEAR = 'year',
	HALF_YEAR = 'half-year',
	QUARTER = 'quarter',
	MONTH = 'month',
	HALF_MONTH = 'half-month',
	TEN_DAYS = 'ten-days',
	WEEK_OF_YEAR = 'week-of-year',
	WEEK_OF_MONTH = 'week-of-month',
	HALF_WEEK = 'half-week',
	DAY_OF_MONTH = 'day-of-month',
	DAY_OF_WEEK = 'day-of-week',
	DAY_KIND = 'day-kind',
	HOUR = 'hour',
	HOUR_KIND = 'hour-kind',
	AM_PM = 'am-pm',


class InspectionTimeRange(DataModel, BaseModel):
	type: InspectionTimeRangeType = None
	value: Any = None


class InspectionYearRange(InspectionTimeRange):
	type: InspectionTimeRangeType.YEAR
	value: int


class InspectionHalfYearRange(InspectionTimeRange):
	type: InspectionTimeRangeType.HALF_YEAR
	value: Literal[1, 2]


class InspectionQuarterRange(InspectionTimeRange):
	type: InspectionTimeRangeType.QUARTER
	value: Literal[1, 2, 3, 4]


class InspectionMonthRange(InspectionTimeRange):
	type: InspectionTimeRangeType.MONTH
	value: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]


class InspectionHalfMonthRange(InspectionTimeRange):
	type: InspectionTimeRangeType.HALF_MONTH
	value: Literal[1, 2]


class InspectionTenDaysRange(InspectionTimeRange):
	type: InspectionTimeRangeType.TEN_DAYS
	value: Literal[1, 2, 3]


class InspectionWeekOfYearRange(InspectionTimeRange):
	type: InspectionTimeRangeType.WEEK_OF_YEAR
	value: int  # 0 - 53


class InspectionWeekOfMonthRange(InspectionTimeRange):
	type: InspectionTimeRangeType.WEEK_OF_MONTH
	value: Literal[0, 1, 2, 3, 4, 5]


class InspectionHalfWeekRange(InspectionTimeRange):
	type: InspectionTimeRangeType.HALF_WEEK
	value: Literal[1, 2]


class InspectionDayOfMonthRange(InspectionTimeRange):
	type: InspectionTimeRangeType.DAY_OF_MONTH
	value: int  # 1 - 31


class InspectionDayOfWeekRange(InspectionTimeRange):
	type: InspectionTimeRangeType.DAY_OF_WEEK
	value: Literal[1, 2, 3, 4, 5, 6, 7]


class InspectionDayKindRange(InspectionTimeRange):
	type: InspectionTimeRangeType.DAY_KIND
	value: Literal[1, 2, 3]


class InspectionHourRange(InspectionTimeRange):
	type: InspectionTimeRangeType.HOUR
	value: int  # / 0 - 23


class InspectionHourKindRange(InspectionTimeRange):
	type: InspectionTimeRangeType.HOUR_KIND
	value: Literal[1, 2, 3]


class InspectionAmPmRange(InspectionTimeRange):
	type: InspectionTimeRangeType.AM_PM
	value: Literal[1, 2]


def construct_time_range(a_range: Optional[Union[dict, InspectionTimeRange]]) -> Optional[InspectionTimeRange]:
	if a_range is None:
		return None
	elif isinstance(a_range, InspectionTimeRange):
		return a_range

	range_type = a_range.get('type')
	if range_type == InspectionTimeRangeType.YEAR:
		return InspectionYearRange(**a_range)
	elif range_type == InspectionTimeRangeType.HALF_YEAR:
		return InspectionHalfYearRange(**a_range)
	elif range_type == InspectionTimeRangeType.QUARTER:
		return InspectionQuarterRange(**a_range)
	elif range_type == InspectionTimeRangeType.MONTH:
		return InspectionMonthRange(**a_range)
	elif range_type == InspectionTimeRangeType.HALF_MONTH:
		return InspectionHalfMonthRange(**a_range)
	elif range_type == InspectionTimeRangeType.TEN_DAYS:
		return InspectionTenDaysRange(**a_range)
	elif range_type == InspectionTimeRangeType.WEEK_OF_YEAR:
		return InspectionWeekOfYearRange(**a_range)
	elif range_type == InspectionTimeRangeType.WEEK_OF_MONTH:
		return InspectionWeekOfMonthRange(**a_range)
	elif range_type == InspectionTimeRangeType.HALF_WEEK:
		return InspectionHalfWeekRange(**a_range)
	elif range_type == InspectionTimeRangeType.DAY_OF_MONTH:
		return InspectionDayOfMonthRange(**a_range)
	elif range_type == InspectionTimeRangeType.DAY_OF_WEEK:
		return InspectionDayOfWeekRange(**a_range)
	elif range_type == InspectionTimeRangeType.DAY_KIND:
		return InspectionDayKindRange(**a_range)
	elif range_type == InspectionTimeRangeType.HOUR:
		return InspectionHourRange(**a_range)
	elif range_type == InspectionTimeRangeType.HOUR_KIND:
		return InspectionHourKindRange(**a_range)
	elif range_type == InspectionTimeRangeType.AM_PM:
		return InspectionAmPmRange(**a_range)
	else:
		raise Exception(f'Inspection time range type[{range_type}] cannot be recognized.')


def construct_time_ranges(ranges: Optional[list] = None) -> Optional[List[InspectionTimeRange]]:
	if ranges is None:
		return None
	else:
		return ArrayHelper(ranges).map(lambda x: construct_time_range(x)).to_list()


class Inspection(UserBasedTuple, Auditable, BaseModel):
	inspectionId: InspectionId = None
	name: str = None
	indicatorId: IndicatorId = None
	# indicator value aggregate arithmetic
	aggregateArithmetics: List[IndicatorAggregateArithmetic] = []
	# none, measure on indicator value or other factor
	measureOn: InspectMeasureOn = None
	# if measure on factor, factor id must be given
	measureOnFactorId: FactorId = None
	# bucket for any measure on type, or no bucket also allowed if measure on factor rather than indicator value
	measureOnBucketId: BucketId = None
	# time range
	timeRangeMeasure: MeasureMethod = None
	# time range factor
	timeRangeFactorId: FactorId = None
	# ranges on time factor for filter data
	timeRanges: List[InspectionTimeRange] = None
	# time measure on factor. measure can use another factor or just measure on the same time factor
	measureOnTime: MeasureMethod = None
	# time measure on factor
	measureOnTimeFactorId: FactorId = None
	criteria: List[IndicatorCriteria] = []

	def __setattr__(self, name, value):
		if name == 'timeRanges':
			super().__setattr__(name, construct_time_ranges(value))
		elif name == 'criteria':
			super().__setattr__(name, construct_indicator_criteria_list(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import DataModel, DataSourceId, OptimisticLock, TenantBasedTuple
from watchmen_utilities import ArrayHelper


class DataSourceParam(DataModel, BaseModel):
	name: str = None
	value: str = None


class DataSourceType(str, Enum):
	MYSQL = 'mysql',
	ORACLE = 'oracle',
	MONGODB = 'mongodb',
	MSSQL = 'mssql',
	POSTGRESQL = 'postgresql',
	OSS = 'oss',
	S3 = 's3'


def construct_param(param: Optional[Union[dict, DataSourceParam]]) -> Optional[DataSourceParam]:
	if param is None:
		return None
	elif isinstance(param, DataSourceParam):
		return param
	else:
		return DataSourceParam(**param)


def construct_params(params: Optional[list] = None) -> Optional[List[DataSourceParam]]:
	if params is None:
		return None
	else:
		return ArrayHelper(params).map(lambda x: construct_param(x)).to_list()


class DataSource(TenantBasedTuple, OptimisticLock, BaseModel):
	dataSourceId: DataSourceId = None
	dataSourceCode: str = None
	dataSourceType: DataSourceType = None
	host: str = None
	port: str = None
	username: str = None
	password: str = None
	name: str = None
	url: str = None
	params: List[DataSourceParam] = []

	def __setattr__(self, name, value):
		if name == 'params':
			super().__setattr__(name, construct_params(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import TypeVar

ChartColor = TypeVar("ChartColor", bound=str)


class ChartType(str, Enum):
	COUNT = 'count'
	BAR = 'bar'
	LINE = 'line'
	SCATTER = 'scatter'
	PIE = 'pie'
	DOUGHNUT = 'doughnut'
	NIGHTINGALE = 'nightingale'
	SUNBURST = 'sunburst'
	TREE = 'tree'
	TREEMAP = 'treemap'
	MAP = 'map'
	CUSTOMIZED = 'customized'



from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import construct_parameter, construct_parameter_joint, FactorId, Parameter
from watchmen_utilities import ArrayHelper
from .pipeline_action import AggregateArithmeticHolder, FindBy, PipelineAction, ToFactor, ToTopic, WriteTopicActionType


class MappingFactor(AggregateArithmeticHolder, BaseModel):
	source: Parameter = None
	factorId: FactorId = None

	def __setattr__(self, name, value):
		if name == 'source':
			super().__setattr__(name, construct_parameter(value))
		else:
			super().__setattr__(name, value)


class MappingRow(PipelineAction):
	mapping: List[MappingFactor] = []


class WriteTopicAction(ToTopic):
	type: WriteTopicActionType = None


def construct_mapping_factor(condition: Optional[Union[dict, MappingFactor]]) -> Optional[MappingFactor]:
	if condition is None:
		return None
	elif isinstance(condition, MappingFactor):
		return condition
	else:
		return MappingFactor(**condition)


def construct_mapping(conditions: Optional[list]) -> Optional[List[MappingFactor]]:
	if conditions is None:
		return None
	return ArrayHelper(conditions).map(lambda x: construct_mapping_factor(x)).to_list()


class InsertRowAction(WriteTopicAction, MappingRow):
	type: WriteTopicActionType = WriteTopicActionType.INSERT_ROW

	def __setattr__(self, name, value):
		if name == 'mapping':
			super().__setattr__(name, construct_mapping(value))
		else:
			super().__setattr__(name, value)


class InsertOrMergeRowAction(WriteTopicAction, MappingRow, FindBy):
	type: WriteTopicActionType = WriteTopicActionType.INSERT_OR_MERGE_ROW

	def __setattr__(self, name, value):
		if name == 'mapping':
			super().__setattr__(name, construct_mapping(value))
		elif name == 'by':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


class MergeRowAction(WriteTopicAction, MappingRow, FindBy):
	type: WriteTopicActionType = WriteTopicActionType.MERGE_ROW

	def __setattr__(self, name, value):
		if name == 'mapping':
			super().__setattr__(name, construct_mapping(value))
		elif name == 'by':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


class WriteFactorAction(ToFactor, WriteTopicAction, FindBy, AggregateArithmeticHolder):
	type: WriteTopicActionType = WriteTopicActionType.WRITE_FACTOR
	source: Parameter = None

	def __setattr__(self, name, value):
		if name == 'mapping':
			super().__setattr__(name, construct_mapping(value))
		elif name == 'by':
			super().__setattr__(name, construct_parameter_joint(value))
		elif name == 'source':
			super().__setattr__(name, construct_parameter(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import Union

from pydantic import BaseModel

from watchmen_model.common import DataModel, FactorId, ParameterJoint, PipelineActionId, TopicId


class SystemActionType(str, Enum):
	ALARM = 'alarm',
	COPY_TO_MEMORY = 'copy-to-memory',
	WRITE_TO_EXTERNAL = 'write-to-external'


class ReadTopicActionType(str, Enum):
	READ_ROW = 'read-row',
	READ_FACTOR = 'read-factor',
	EXISTS = 'exists',
	READ_ROWS = 'read-rows',
	READ_FACTORS = 'read-factors'


class WriteTopicActionType(str, Enum):
	MERGE_ROW = 'merge-row',
	INSERT_ROW = 'insert-row',
	INSERT_OR_MERGE_ROW = 'insert-or-merge-row',
	WRITE_FACTOR = 'write-factor'


class DeleteTopicActionType(str, Enum):
	DELETE_ROW = 'delete-row',
	DELETE_ROWS = 'delete-rows'


PipelineActionType = Union[WriteTopicActionType, ReadTopicActionType, DeleteTopicActionType, SystemActionType]


class PipelineAction(DataModel, BaseModel):
	actionId: PipelineActionId = None
	type: PipelineActionType = None


class MemoryWriter(PipelineAction):
	variableName: str = None


class FromTopic(PipelineAction):
	topicId: TopicId = None


class FromFactor(FromTopic):
	factorId: FactorId = None


class ToTopic(PipelineAction):
	topicId: TopicId = None


class ToFactor(ToTopic):
	factorId: FactorId = None


class FindBy(PipelineAction):
	by: ParameterJoint = None


class AggregateArithmetic(str, Enum):
	NONE = 'none',
	COUNT = 'count',
	SUM = 'sum',
	AVG = 'avg'


class AggregateArithmeticHolder(DataModel):
	arithmetic: AggregateArithmetic = None



from typing import List

from pydantic import BaseModel

from watchmen_model.common import ConnectedSpaceId, DashboardId, LastVisit, UserBasedTuple


class Favorite(UserBasedTuple, LastVisit, BaseModel):
	connectedSpaceIds: List[ConnectedSpaceId] = []
	dashboardIds: List[DashboardId] = []



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import BucketId, construct_parameter_joint, DataModel, FactorId, IndicatorId, \
	OptimisticLock, ParameterJoint, SubjectDatasetColumnId, SubjectId, TenantBasedTuple, TopicId, UserGroupId
from watchmen_utilities import ArrayHelper


class IndicatorAggregateArithmetic(str, Enum):
	COUNT = 'count'
	SUM = 'sum'
	AVG = 'avg'
	MAX = 'max'
	MIN = 'min'


class RelevantIndicatorType(str, Enum):
	SAME = 'same'
	HIGH_CORRELATED = 'high-correlated'
	WEAK_CORRELATED = 'weak-correlated'
	# /** this causes relevant */
	THIS_CAUSES_RELEVANT = 'this-causes-relevant'
	# /** relevant causes this */
	RELEVANT_CAUSES_THIS = 'relevant-causes-this'


class RelevantIndicator(DataModel, BaseModel):
	indicatorId: IndicatorId = None
	type: RelevantIndicatorType = None


def construct_relevant(relevant: Optional[Union[dict, RelevantIndicator]]) -> Optional[RelevantIndicator]:
	if relevant is None:
		return None
	elif isinstance(relevant, RelevantIndicator):
		return relevant
	else:
		# noinspection PyArgumentList
		return RelevantIndicator(**relevant)


def construct_relevants(relevants: Optional[list] = None) -> Optional[List[RelevantIndicator]]:
	if relevants is None:
		return None
	else:
		return ArrayHelper(relevants).map(lambda x: construct_relevant(x)).to_list()


class IndicatorBaseOn(str, Enum):
	TOPIC = 'topic',
	SUBJECT = 'subject'


# noinspection PyRedundantParentheses,DuplicatedCode
class AvoidFastApiError:
	joint: ParameterJoint = None


class IndicatorFilter(DataModel, AvoidFastApiError, BaseModel):
	enabled: bool = False

	def __setattr__(self, name, value):
		if name == 'joint':
			super().__setattr__(name, construct_parameter_joint(value))
		else:
			super().__setattr__(name, value)


def construct_filter(a_filter: Optional[Union[dict, IndicatorFilter]]) -> Optional[IndicatorFilter]:
	if a_filter is None:
		return None
	elif isinstance(a_filter, IndicatorFilter):
		return a_filter
	else:
		# noinspection PyArgumentList
		return IndicatorFilter(**a_filter)


class Indicator(TenantBasedTuple, OptimisticLock, BaseModel):
	indicatorId: IndicatorId = None
	name: str = None
	# when indicator is on topic
	topicOrSubjectId: Union[TopicId, SubjectId] = None
	# is a count indicator when factor is not appointed
	# it is columnId when base one a subject
	factorId: Union[FactorId, SubjectDatasetColumnId] = None
	baseOn: IndicatorBaseOn = None
	category1: str = None
	category2: str = None
	category3: str = None
	description: str = None
	# effective only when factorId is appointed
	valueBuckets: List[BucketId] = []
	# noinspection SpellCheckingInspection
	relevants: List[RelevantIndicator] = []
	groupIds: List[UserGroupId] = None
	filter: IndicatorFilter = None

	def __setattr__(self, name, value):
		if name == 'relevants':
			super().__setattr__(name, construct_relevants(value))
		elif name == 'filter':
			super().__setattr__(name, construct_filter(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import Auditable, ConnectedSpaceId, construct_parameter, construct_parameter_conditions, \
	construct_parameter_joint, DataModel, FactorId, LastVisit, Pageable, Parameter, ParameterCondition, \
	ParameterJoint, SubjectDatasetColumnId, SubjectId, TopicId, UserBasedTuple
from watchmen_utilities import ArrayHelper


class SubjectJoinType(str, Enum):
	LEFT = 'left',
	RIGHT = 'right',
	INNER = 'inner',


class SubjectDatasetJoin(DataModel, BaseModel):
	topicId: TopicId = None
	factorId: FactorId = None
	secondaryTopicId: TopicId = None
	secondaryFactorId: FactorId = None
	type: SubjectJoinType = SubjectJoinType.INNER


class SubjectColumnArithmetic(str, Enum):
	NONE = 'none'
	COUNT = 'count'
	SUMMARY = 'sum'
	AVERAGE = 'avg'
	MAXIMUM = 'max'
	MINIMUM = 'min'


class SubjectDatasetColumn(DataModel, BaseModel):
	columnId: SubjectDatasetColumnId = None
	parameter: Parameter
	alias: str = None
	arithmetic: SubjectColumnArithmetic = None

	def __setattr__(self, name, value):
		if name == 'parameter':
			super().__setattr__(name, construct_parameter(value))
		else:
			super().__setattr__(name, value)


def construct_column(column: Optional[Union[dict, SubjectDatasetColumn]]) -> Optional[SubjectDatasetColumn]:
	if column is None:
		return None
	elif isinstance(column, SubjectDatasetColumn):
		return column
	else:
		return SubjectDatasetColumn(**column)


def construct_columns(columns: Optional[list] = None) -> Optional[List[SubjectDatasetColumn]]:
	if columns is None:
		return None
	else:
		return ArrayHelper(columns).map(lambda x: construct_column(x)).to_list()


def construct_join(join: Optional[Union[dict, SubjectDatasetJoin]]) -> Optional[SubjectDatasetJoin]:
	if join is None:
		return None
	elif isinstance(join, SubjectDatasetJoin):
		return join
	else:
		return SubjectDatasetJoin(**join)


def construct_joins(joins: Optional[list] = None) -> Optional[List[SubjectDatasetJoin]]:
	if joins is None:
		return None
	else:
		return ArrayHelper(joins).map(lambda x: construct_join(x)).to_list()


class AvoidFastApiError:
	filters: ParameterJoint


class SubjectDataset(DataModel, AvoidFastApiError, BaseModel):
	columns: List[SubjectDatasetColumn] = []
	joins: List[SubjectDatasetJoin] = []

	def __setattr__(self, name, value):
		if name == 'filters':
			super().__setattr__(name, construct_parameter_joint(value))
		elif name == 'columns':
			super().__setattr__(name, construct_columns(value))
		elif name == 'joins':
			super().__setattr__(name, construct_joins(value))
		else:
			super().__setattr__(name, value)


def construct_dataset(dataset: Optional[dict] = None) -> Optional[SubjectDataset]:
	if dataset is None:
		return None
	elif isinstance(dataset, SubjectDataset):
		return dataset
	else:
		return SubjectDataset(**dataset)


class Subject(UserBasedTuple, Auditable, LastVisit, BaseModel):
	subjectId: SubjectId = None
	name: str = None
	connectId: ConnectedSpaceId = None
	autoRefreshInterval: Optional[int] = 0
	dataset: SubjectDataset = None

	def __setattr__(self, name, value):
		if name == 'dataset':
			super().__setattr__(name, construct_dataset(value))
		else:
			super().__setattr__(name, value)


class SubjectDatasetCriteriaIndicatorArithmetic(str, Enum):
	NONE = 'none'
	COUNT = 'count'
	SUMMARY = 'sum'
	AVERAGE = 'avg'
	MAXIMUM = 'max'
	MINIMUM = 'min'


class SubjectDatasetCriteriaIndicator(BaseModel):
	name: str = None
	arithmetic: SubjectDatasetCriteriaIndicatorArithmetic = None
	alias: str = None


def construct_criteria_indicator(
		indicator: Optional[Union[dict, SubjectDatasetCriteriaIndicator]]) -> Optional[SubjectDatasetCriteriaIndicator]:
	if indicator is None:
		return None
	elif isinstance(indicator, SubjectDatasetCriteriaIndicator):
		return indicator
	else:
		return SubjectDatasetCriteriaIndicator(**indicator)


def construct_criteria_indicators(indicators: Optional[list]) -> Optional[List[SubjectDatasetCriteriaIndicator]]:
	if indicators is None:
		return None
	return ArrayHelper(indicators).map(lambda x: construct_criteria_indicator(x)).to_list()


class SubjectDatasetCriteria(Pageable):
	# use one of subject id or name
	subjectId: Optional[SubjectId]
	subjectName: Optional[str]
	indicators: List[SubjectDatasetCriteriaIndicator]
	conditions: List[ParameterCondition]

	def __setattr__(self, name, value):
		if name == 'indicators':
			super().__setattr__(name, construct_criteria_indicators(value))
		elif name == 'conditions':
			super().__setattr__(name, construct_parameter_conditions(value))
		else:
			super().__setattr__(name, value)



from enum import Enum
from typing import Optional, TypeVar, Union

from pydantic import BaseModel

from watchmen_model.common import DataModel, FactorId, TenantBasedTuple, TopicId

MonitorRuleId = TypeVar('MonitorRuleId', bound=str)


class MonitorRuleCode(str, Enum):
	# structure
	RAW_MISMATCH_STRUCTURE = 'raw-mismatch-structure',

	# type
	FACTOR_MISMATCH_ENUM = 'factor-mismatch-enum',
	FACTOR_MISMATCH_TYPE = 'factor-mismatch-type',
	FACTOR_MISMATCH_DATE_TYPE = 'factor-mismatch-date-type',

	# topic row count
	ROWS_NOT_EXISTS = 'rows-not-exists',
	ROWS_NO_CHANGE = 'rows-no-change',
	ROWS_COUNT_MISMATCH_AND_ANOTHER = 'rows-count-mismatch-and-another',

	# for all factor types
	FACTOR_IS_EMPTY = 'factor-is-empty',
	FACTOR_USE_CAST = 'factor-use-cast',
	FACTOR_COMMON_VALUE_OVER_COVERAGE = 'factor-common-value-over-coverage',
	FACTOR_EMPTY_OVER_COVERAGE = 'factor-empty-over-coverage',

	# for number type
	FACTOR_BREAKS_MONOTONE_INCREASING = 'factor-breaks-monotone-increasing',
	FACTOR_BREAKS_MONOTONE_DECREASING = 'factor-breaks-monotone-decreasing',
	FACTOR_NOT_IN_RANGE = 'factor-not-in-range',
	FACTOR_MAX_NOT_IN_RANGE = 'factor-max-not-in-range',
	FACTOR_MIN_NOT_IN_RANGE = 'factor-min-not-in-range',
	FACTOR_AVG_NOT_IN_RANGE = 'factor-avg-not-in-range',
	FACTOR_MEDIAN_NOT_IN_RANGE = 'factor-median-not-in-range',
	FACTOR_QUANTILE_NOT_IN_RANGE = 'factor-quantile-not-in-range',
	# noinspection SpellCheckingInspection
	FACTOR_STDEV_NOT_IN_RANGE = 'factor-stdev-not-in-range',
	FACTOR_COMMON_VALUE_NOT_IN_RANGE = 'factor-common-value-not-in-range',

	# for string type
	FACTOR_IS_BLANK = 'factor-is-blank',
	FACTOR_STRING_LENGTH_MISMATCH = 'factor-string-length-mismatch',
	FACTOR_STRING_LENGTH_NOT_IN_RANGE = 'factor-string-length-not-in-range',
	FACTOR_MATCH_REGEXP = 'factor-match-regexp',
	FACTOR_MISMATCH_REGEXP = 'factor-mismatch-regexp',

	# for 2 factors
	FACTOR_AND_ANOTHER = 'factor-and-another'


class MonitorRuleGrade(str, Enum):
	GLOBAL = 'global',
	TOPIC = 'topic',
	FACTOR = 'factor'


class MonitorRuleSeverity(str, Enum):
	FATAL = 'fatal',
	WARN = 'warn',
	TRACE = 'trace',


class MonitorRuleStatisticalInterval(str, Enum):
	DAILY = 'daily',
	WEEKLY = 'weekly',
	MONTHLY = 'monthly'


class MonitorRuleCompareOperator(str, Enum):
	EQUAL = 'eq',
	LESS_THAN = 'lt',
	LESS_THAN_OR_EQUAL = 'lte',
	GREATER_THAN = 'gt',
	GREATER_THAN_EQUAL = 'gte'


class MonitorRuleParameters(DataModel, BaseModel):
	statisticalInterval: MonitorRuleStatisticalInterval = None
	coverageRate: int = None
	aggregation: int = None
	quantile: int = None
	length: int = None
	max: int = None
	min: int = None
	regexp: str = None
	compareOperator: MonitorRuleCompareOperator = None
	topicId: Optional[TopicId] = None
	factorId: Optional[FactorId] = None


def construct_params(params: Optional[Union[dict, MonitorRuleParameters]]) -> Optional[MonitorRuleParameters]:
	if params is None:
		return None
	elif isinstance(params, MonitorRuleParameters):
		return params
	else:
		return MonitorRuleParameters(**params)


class MonitorRule(TenantBasedTuple, BaseModel):
	ruleId: MonitorRuleId = None
	code: MonitorRuleCode = None
	grade: MonitorRuleGrade = None
	severity: MonitorRuleSeverity = None
	topicId: TopicId = None
	factorId: Optional[FactorId] = None
	params: MonitorRuleParameters = None
	enabled: bool = False

	def __setattr__(self, name, value):
		if name == 'params':
			super().__setattr__(name, construct_params(value))
		else:
			super().__setattr__(name, value)



from copy import deepcopy
from datetime import date, time
from decimal import Decimal
from typing import Any, Dict

from pydantic import BaseModel


def copy_x(x: Any) -> Any:
	if x is None:
		return x
	elif isinstance(x, dict):
		return copy_dict(x)
	elif isinstance(x, list):
		return copy_list(x)
	elif isinstance(x, tuple):
		return copy_tuple(x)
	elif isinstance(x, (str, int, float, Decimal, bool, date, time)):
		return deepcopy(x)
	elif isinstance(x, BaseModel):
		return x.dict()
	elif isinstance(x, DataModel):
		return x.to_dict()
	else:
		raise ValueError(f'X[{x}] cannot be copied.')


def copy_tuple(a_tuple: tuple) -> tuple:
	new_list = []
	for v in a_tuple:
		new_list.append(copy_x(v))
	return tuple(new_list)


def copy_list(a_list: list) -> list:
	new_list = []
	for v in a_list:
		new_list.append(copy_x(v))
	return new_list


def copy_dict(a_dict: dict) -> dict:
	new_dict = {}
	for k, v in a_dict.items():
		new_dict[copy_x(k)] = copy_x(v)
	return new_dict


class DataModel:
	def __init__(self, **data: Any):
		for key, value in data.items():
			self.__setattr__(key, value)

	def __setattr__(self, name, value):
		self.__dict__[name] = value

	def __getattr__(self, name) -> Any:
		# to avoid property not found
		return self.__dict__.get(name)

	def to_dict(self) -> Dict[str, Any]:
		return copy_dict(self.__dict__)


class SettingsModel(DataModel):
	pass
